"""
Global constants
Stores paths to directories used for input files, as well as constants referenced throughout the code base
"""
from pathlib import Path

from fastsim import vehicle
import os
# ./t3co
wkdir = Path(__file__).parent.parent

SWEEP_PATH = Path(os.path.abspath(__file__)).parents[1]/"sweep.py"

### modified from 1/2/3/4: conv/hev/phev/bev to match FASTSim strings

CONV = vehicle.CONV
HEV = vehicle.HEV
PHEV = vehicle.PHEV
BEV = vehicle.BEV
PT_TYPES_NUM_TO_STR = {CONV: "Conv", HEV: "HEV", PHEV: "PHEV", BEV: "BEV"}
###

FC_EFF_TYPES = {
    1: "SI",
    2: "Diesel - ISB280",
    3: "Diesel",
    4: "Fuel Cell",
    5: "Hybrid Diesel",
    6: "Diesel - HD",
    7: "Diesel - HDISM Scaled",
    8: "Diesel - HDISM Scaled",
    9: "CNG",
}
# note that FC_EFF_TYPES comes from fastsim and this here does not seem to override. Need to make sure it matches, or perhaps read it from fastsim.vehicle

STD_VAR_NAMES = "stdVarNames"

maxGvwrKg = 0
evGVWRAllowanceLbs = 0

# store reference to range cycle used for optimization
global_opt_range_cycle = None

# store scenario_name of current run
vocation_scenario = None

kwh_per_gge = 33.7

DieselGalPerGasGal = 0.887  # energy equivalent gallons of diesel per 1 gallon gas

kgH2_per_gge = 1.019  # https://epact.energy.gov/fuel-conversion-factors for Hydrogen

mps_to_mph = 2.23694  # 1 mps = 2.23694 mph
m_to_mi = 0.000621371  # 1 m = 0.000621371 mi


def get_kwh_per_gge():
    """
    This is a getter for kwh_per_gge, sim and scenario dependant var that can change
    important to get from one location each time so we can track when and how it's used

    Returns:
        kwh_per_gge (float): kWh per Gasoline Gallon Equivalent
    """
    return kwh_per_gge


# --------------------------- ###  directories and files ### ----------------------------


TCO_INTERMEDIATES = None
TCO_RESULTS = None
TCO_RES_FIGS = None

#  ## resources

# ./t3co/resources
OPTIMIZATION_AND_TCO_RCRS = Path(os.path.abspath(__file__)).parents[1] / "resources"

# ./t3co/resources/cycles
OPTIMIZATION_DRIVE_CYCLES = OPTIMIZATION_AND_TCO_RCRS / "cycles"

# benchmark resources
T2COBENCHMARKDATADIR = OPTIMIZATION_AND_TCO_RCRS / "benchmarkdata"

T3CO_INPUTS_DIR = OPTIMIZATION_AND_TCO_RCRS / "inputs"

def set_tco_intermediates():
    """
    This function sets path for TCO_INTERMEDIATES to save tsv files
    """
    global TCO_INTERMEDIATES

    # ./t3co/resources/f'vehicles/{vocation_scenario}/tco/tco_intermediates'
    TCO_INTERMEDIATES = (
        OPTIMIZATION_AND_TCO_RCRS
        / f"vehicles/{vocation_scenario}/tco/tco_intermediates"
    )
    if not TCO_INTERMEDIATES.exists():
        TCO_INTERMEDIATES.mkdir(parents=True)


def set_tco_results():
    """
    This function sets path for TCO_RESULTS
    """
    global TCO_RESULTS
    # ./t3co/resources/f'vehicles/{vocation_scenario}/tco/tco_results'
    TCO_RESULTS = (
        OPTIMIZATION_AND_TCO_RCRS / f"vehicles/{vocation_scenario}/tco/tco_results"
    )
    if not TCO_RESULTS.exists():
        TCO_RESULTS.mkdir(parents=True)

    global TCO_RES_FIGS
    # ./t3co/resources/f'vehicles/{vocation_scenario}/result_figures/'
    TCO_RES_FIGS = (
        OPTIMIZATION_AND_TCO_RCRS / f"vehicles/{vocation_scenario}/result_figures/"
    )
    if not TCO_RES_FIGS.exists():
        TCO_RES_FIGS.mkdir(parents=True)


OPTIMIZATION_RESOURCES_AUX = OPTIMIZATION_AND_TCO_RCRS / "auxiliary"
# FASTSim and Scenario input files
FASTSIM_INPUTS_FILE = "FASTSimInputsHeader.csv"
OTHER_INPUTS_FILE = "OtherInputs.csv"

# ./t3co/resources/"FASTSimInputs.csv"
FASTSIM_INPUTS = OPTIMIZATION_RESOURCES_AUX / FASTSIM_INPUTS_FILE

# ./t3co/resources/"OtherInputs.csv"
OTHER_INPUTS = OPTIMIZATION_RESOURCES_AUX / OTHER_INPUTS_FILE

# ./t3co/resources/'FuelPrices.csv'
REGIONAL_FUEL_PRICES_BY_TYPE_BY_YEAR = OPTIMIZATION_RESOURCES_AUX / "FuelPrices.csv"

RESIDUAL_VALUE_PER_YEAR = OPTIMIZATION_RESOURCES_AUX / "ResidualValues.csv"
MOO_KNOB_SWEEP_PLOTS_DIR = Path(os.path.abspath(__file__)).parents[1]/"tco_results" / "knob_sweep_results"

# TCO input files
write_files = False
ANN_TRAVEL_TSV = "annual-travel.tsv"
EMISSION_RATE_TSV = "emission-rate.tsv"
FUEL_EFF_TSV = "fuel-efficiency.tsv"
FUEL_EXPENSE_TSV = "fuel-expense.tsv"
FUEL_SPLIT_TSV = "fuel-split.tsv"
MARKET_SHARE_TSV = "market-share.tsv"
REGIONAL_SALES_TSV = "regional-sales.tsv"
SURVIVAL_TSV = "survival.tsv"
TRAVEL_EXP_TSV = "travel-expense.tsv"
VEH_EXP_TSV = "vehicle-expense.tsv"

# T3CO modules
TCO = "tests"

# Testing dirs

# ./t3co/tests
TESTSDIR = Path(os.path.abspath(__file__)).parents[1] / f"{TCO}/tco_tests"

# ./t3co/tests/tco_tests/'test_cycles'
TESTCYCLES = TESTSDIR / "test_cycles"

# ./t3co/tests/tco_tests/'test_vehicles'
TESTVEHICLES = TESTSDIR / "test_vehicles"

TESTVEHICLEINPUTS = TESTSDIR / "TCO_VEHICLE_TEST_INPUTS.csv"
TESTSCENARIOINPUTS = TESTSDIR / "TCO_SCENARIO_TEST_INPUTS.csv"

# ## useful functions

KG_2_LB = 2.20462


def kg_to_lbs(kgs:float)->float:
    """
    This function converts kg to lb

    Args:
        kgs (float): mass in kg

    Returns:
        (float): mass in pounds
    """
    return kgs * KG_2_LB


def lbs_to_kgs(lbs:float)->float:
    """
    This function converts lb to kg

    Args:
        lbs (float): mass in pounds

    Returns:
        (float): mass in kg
    """
    return lbs / KG_2_LB


def not_falsy(var:float)->bool:
    """
    This function returns True to verify that var is NOT falsy: not in [None, np.nan, 0, False]


    Args:
        var (float): variable to check

    Returns:
        (bool): True if not in [None, 0, False]
    """
    return var not in [None, 0, False]



from __future__ import annotations

import numpy as np
import pandas as pd

from t3co.run import Global as gl
from t3co.run import run_scenario
from t3co.tco import opportunity_cost
import fastsim

KG_2_LB = 2.20462


def kg_to_lbs(kgs: float) -> float:
    return kgs * KG_2_LB


# keeping this for when we do Emissions work
# with open(gl.TCO_INTERMEDIATES / gl.EMISSION_RATE_TSV, 'w', newline='') as er_file:
#         writer = csv.writer(er_file, delimiter='\t')
#         writer.writerow(["Vehicle", "Model Year", "Fuel", "Age [yr]", "Region", "Pollutant", "Emission Rate [g/gge]", "Vocation" ])


def find_residual_rates(
    vehicle: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario
) -> float:  # finds residual rate at end of vehicle life
    """
    This helper method gets the residual rates from ResidualValues.csv

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of analysis vehicle
        scenario (run_scenario.Scenario): Scenario object of current selection

    Returns:
        residual_rates (float): Residual rate as percentage of MSRP
    """
    residual_rates_all = pd.read_csv(gl.RESIDUAL_VALUE_PER_YEAR)
    vehicle_class = scenario.vehicle_class
    powertrain_type = vehicle.veh_pt_type.lower()
    year = str(scenario.vehicle_life_yr)
    residual_rates = residual_rates_all.loc[
        (residual_rates_all["VehicleClass"].str.lower() == vehicle_class)
        & (residual_rates_all["PowertrainType"].str.lower() == powertrain_type)
    ][year].values[0]
    return residual_rates


def calculate_dollar_cost(
    veh: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario
) -> dict:
    """
    This helper method calculates the MSRP breakdown dictionary from
    -   Glider
    -   Fuel converter
    -   Fuel Storage
    -   Motor & power electronics
    -   Plug
    -   Battery
    -   Battery replacement
    -   Purchase tax

    Args:
        veh (fastsim.vehicle.Vehicle): FASTSim vehicle object of analysis vehicle
        scenario (run_scenario.Scenario): Scenario object of current selection

    Returns:
        cost_set (dict): Dictionary containing MSRP breakdown
    """
    chargingOn = False

    fc_ice_base_cost_dol = scenario.fc_ice_base_cost_dol
    markup_pct = scenario.markup_pct
    fc_ice_cost_dol_per_kw = scenario.fc_ice_cost_dol_per_kw
    pe_mc_base_cost_dol = scenario.pe_mc_base_cost_dol
    pe_mc_cost_dol_per_kw = scenario.pe_mc_cost_dol_per_kw
    ess_base_cost_dol = scenario.ess_base_cost_dol
    ess_cost_dol_per_kwh = scenario.ess_cost_dol_per_kwh
    plugPrice = scenario.plug_base_cost_dol
    tax_rate_pct = scenario.tax_rate_pct
    # TODO add fc_fuelcell_cost_dol_per_kw in this list?

    # vehicle glider price does not have markup_pct applied to it
    # on 11/23/2021 it was discussed but we decided it's fine to keep it this way. Glider price is an input that could
    # already have this factored in, especially for TDA, the current main user of T3CO
    vehicle_glider_cost_dol = scenario.vehicle_glider_cost_dol

    veh_pt_type = veh.veh_pt_type

    fc_max_kw = veh.fc_max_kw

    # old python fcEffType maps
    # if self.fcEffType == 1:    # SI:         SI engine
    # elif self.fcEffType == 2:  # Atkinson:   Atkinson cycle SI engine -- greater expansion
    # elif self.fcEffType == 3:  # Diesel:     Diesel (compression ignition) engine
    # elif self.fcEffType == 4:  # H2FC:       H2 fuel cell
    # elif self.fcEffType == 5:  # HD_Diesel:  heavy duty Diesel engine
    # new enumeration for types
    # FC_EFF_TYPES = ("SI", "Atkinson", "Diesel", "H2FC", "HD_Diesel")

    # #  TODO, need an assert for veh.fcEffType in vehicle.effTypes []
    # fcPrice
    if veh.veh_pt_type == gl.BEV or veh.fc_max_kw == 0:
        fcPrice = 0

    elif veh.fc_eff_type == "H2FC":
        fcPrice = scenario.fc_fuelcell_cost_dol_per_kw * fc_max_kw
    # TODO, what should 9 map too??
    elif veh.fc_eff_type == 9:
        fcPrice = (
            scenario.fc_cng_ice_cost_dol_per_kw * fc_max_kw
        ) + fc_ice_base_cost_dol

    else:
        fcPrice = (fc_ice_cost_dol_per_kw * fc_max_kw) + fc_ice_base_cost_dol
    fcPrice *= markup_pct

    # TODO,this should get handled in fastsim.vehicle on INPUT VALIDATION
    # that code has a bug and some logic errors, needs fixing
    # if fcPrice != 0:
    #     eff_type_err_msg = f"ERROR: veh.fc_eff_type {veh.fc_eff_type} must be part of vehicle.FC_EFF_TYPES {vehicle.FC_EFF_TYPES}"
    #     eff_type_err_msg += """\n\toverwrite 1 with 'SI'\n\toverwrite 2 with 'Atkinson'"""
    #     eff_type_err_msg += """\n\toverwrite 3 with 'Diesel'\n\toverwrite 4 with 'H2FC'"""
    #     eff_type_err_msg += """\n\toverwrite 5 with 'HD_Diesel'"""
    #     assert veh.fc_eff_type in vehicle.FC_EFF_TYPES, eff_type_err_msg

    # fuelStorPrice
    if veh.veh_pt_type == gl.BEV:
        fuelStorPrice = 0
    elif veh.veh_pt_type == gl.HEV and scenario.fuel_type[0] == "hydrogen":
        fuelStorPrice = scenario.fs_h2_cost_dol_per_kwh * veh.fs_kwh
    elif (
        veh.veh_pt_type in [gl.CONV, gl.HEV, gl.PHEV] and scenario.fuel_type[0] == "cng"
    ):
        fuelStorPrice = scenario.fs_cng_cost_dol_per_kwh * veh.fs_kwh
    elif veh.veh_pt_type in [gl.CONV, gl.HEV, gl.PHEV]:
        fuelStorPrice = scenario.fs_cost_dol_per_kwh * veh.fs_kwh
    fuelStorPrice *= markup_pct

    # calculate mcPrice
    mc_max_kw = veh.mc_max_kw
    if mc_max_kw == 0:
        mcPrice = 0
    else:
        mcPrice = pe_mc_base_cost_dol + (pe_mc_cost_dol_per_kw * mc_max_kw)
    mc_max_kw *= markup_pct

    # calc ESS price
    if veh.ess_max_kwh == 0:
        essPrice = 0
    else:
        essPrice = ess_base_cost_dol + (ess_cost_dol_per_kwh * veh.ess_max_kwh)
    essPrice *= markup_pct

    # calc plugPrice
    if (
        veh_pt_type == gl.PHEV
        or veh_pt_type == gl.BEV
        or (veh_pt_type == gl.HEV and chargingOn)
    ):
        plugPrice = plugPrice
    else:
        plugPrice = 0
    plugPrice *= markup_pct

    if veh_pt_type == gl.CONV:
        msrp = vehicle_glider_cost_dol + fuelStorPrice + fcPrice
    # could be HEV or FCEV
    elif veh_pt_type == gl.HEV:
        msrp = vehicle_glider_cost_dol + fuelStorPrice + fcPrice + mcPrice + essPrice
    elif veh_pt_type == gl.PHEV:
        msrp = (
            vehicle_glider_cost_dol
            + fuelStorPrice
            + fcPrice
            + mcPrice
            + essPrice
            + plugPrice
        )
    elif veh_pt_type == gl.BEV:
        msrp = vehicle_glider_cost_dol + mcPrice + essPrice + plugPrice

    pTaxCost = tax_rate_pct * msrp

    # insurance_cost = calc_insurance_costs(scenario=scenario, msrp=msrp)
    # residual_cost = calc_residual_cost(vehicle=veh, scenario=scenario, msrp=msrp)
    cost_set = {
        "Glider": vehicle_glider_cost_dol,
        "Fuel converter": fcPrice,
        "Fuel Storage": fuelStorPrice,
        "Motor & power electronics": mcPrice,
        "Plug": plugPrice,
        "Battery": essPrice,
        "Battery replacement": 0,
        "Purchase tax": pTaxCost,
        "msrp": msrp,
        # "Insurance": insurance_cost,
        # "Residual Cost": residual_cost
    }
    return cost_set


def calculate_opp_costs(
    vehicle: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario, range_dict: dict
) -> dict:
    """
    This helper method calculates opportunity costs and generates veh_opp_cost_set from
    -   Payload Lost Capacity Cost/Multiplier
    -   Fueling Downtime
    -   Maintenance and Repair Downtime

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of analysis vehicle
        scenario (run_scenario.Scenario): Scenario object of current selection
        range_dict (dict): Dictionary containing range values from fueleconomy.get_range_mi()

    Returns:
        veh_opp_cost_set (dict): Dictionary containing opportunity cost results
    """
    oppcostobj = opportunity_cost.OpportunityCost(scenario, range_dict)
    if scenario.activate_tco_payload_cap_cost_multiplier:
        # assert optvehicle.veh_pt_type in [gl.BEV, gl.HEV], "payload cap loss factor TCO element only available for BEVs and FCEV HEVs"
        assert gl.not_falsy(scenario.plf_ref_veh_empty_mass_kg)
        assert np.isnan(scenario.plf_ref_veh_empty_mass_kg) == False
        oppcostobj.set_payload_loss_factor(vehicle, scenario)

    if scenario.activate_tco_fueling_dwell_time_cost:
        # assert optvehicle.veh_pt_type in [gl.BEV, gl.HEV], "payload cap loss factor TCO element only available for BEVs and FCEV HEVs"
        assert gl.not_falsy(scenario.downtime_oppy_cost_dol_per_hr)
        assert not np.isnan(scenario.fdt_num_free_dwell_trips)
        assert not np.isnan(scenario.fdt_dwpt_fraction_power_pct)
        assert not np.isnan(scenario.fdt_available_freetime_hr)
        assert not np.isnan(scenario.fdt_avg_overhead_hr_per_dwell_hr)
        assert (
            gl.not_falsy(scenario.shifts_per_year)
            and len(oppcostobj.shifts_per_year) >= scenario.vehicle_life_yr
        ), f"Provide scenario.shifts_per_year as a vector of length > scenario.vehicle_life_yr. Currently {len(oppcostobj.shifts_per_year)}"
        assert gl.not_falsy(scenario.fdt_frac_full_charge_bounds)
        oppcostobj.set_fueling_dwell_time_cost(vehicle, scenario)

    if scenario.activate_mr_downtime_cost:
        assert gl.not_falsy(scenario.mr_planned_downtime_hr_per_yr)
        assert any(np.isnan(scenario.mr_unplanned_downtime_hr_per_mi)) == False
        assert np.isnan(scenario.mr_tire_replace_downtime_hr_per_event) == False
        assert gl.not_falsy(scenario.mr_avg_tire_life_mi)
        oppcostobj.set_M_R_downtime_cost(vehicle, scenario)

    # veh_opp_cost_set = {'payload_cap_cost_multiplier' : None, 'net_fueling_dwell_time_hr_per_yr' : 0., 'dwell_time_cost_Dol' : 0.}
    veh_opp_cost_set = {
        "payload_cap_cost_multiplier": oppcostobj.payload_cap_cost_multiplier,
        "daily_trip_distance_mi": oppcostobj.d_trip_mi,
        "net_fueling_dwell_time_hr_per_yr": oppcostobj.net_fueling_dwell_time_hr_per_yr,
        "fueling_dwell_labor_cost_dol_per_yr": oppcostobj.fueling_dwell_labor_cost_dol_per_yr,
        "fueling_downtime_oppy_cost_dol_per_yr": oppcostobj.fueling_downtime_oppy_cost_dol_per_yr,
        "net_mr_downtime_hr_per_yr": oppcostobj.net_net_mr_downtime_hr_per_yr_per_yr,
        "mr_downtime_oppy_cost_dol_per_yr": oppcostobj.mr_downtime_oppy_cost_dol_per_yr,
        "total_downtime_hr_per_yr": np.array(oppcostobj.net_fueling_dwell_time_hr_per_yr) + np.array(oppcostobj.net_net_mr_downtime_hr_per_yr_per_yr),
    }
    return veh_opp_cost_set


def fill_fuel_eff_file(
    vehicle: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario, mpgge_dict: dict
) -> pd.DataFrame:
    """
    This helper method generates a dataframe of Fuel Efficiency [mi/gge]
    For PHEV, cd_grid_electric_mpgge, cd_fuel_mpgge, and cs_fuel_mpgge
    For BEV, grid_mpgge
    For HEV and CONV, mpgge

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of analysis vehicle
        scenario (run_scenario.Scenario): Scenario object of current selection
        mpgge_dict (dict): MPGGE dictionary from fueleconomy.get_mpgge()

    Returns:
        fefdata (pd.DataFrame): Dictionary containing Fuel Efficiency [mi/gge]
    """
    # this fuel efficiency value is for over all fuel costs
    # so it needs to factor in charging efficiency, if applicable

    if vehicle.veh_pt_type == gl.PHEV:
        # grid efficiency charge depleting mpgge first, then charge sustaining
        mpgges = [
            mpgge_dict["cd_grid_electric_mpgge"],
            mpgge_dict["cd_fuel_mpgge"],
            mpgge_dict["cs_fuel_mpgge"],
        ]
    if vehicle.veh_pt_type == gl.BEV:
        mpgges = [mpgge_dict["grid_mpgge"]]
    if vehicle.veh_pt_type in [gl.HEV, gl.CONV]:
        mpgges = [mpgge_dict["mpgge"]]
    vehicle_segment_name = scenario.segment_name
    model_year = int(scenario.model_year)
    region = scenario.region
    vocation = scenario.vocation

    # get age of vehicle
    # if age < 0: age = "*"
    # can ignore age, so always make it "*" - Kevin Bennion, 8/12/2019
    age = "*"

    fuels = scenario.fuel_type

    assert len(fuels) == len(mpgges), f"fuels/mpgges: {fuels}/{mpgges}"
    if (
        vehicle.veh_pt_type == gl.PHEV
    ):  # FIXME: this assumes the PHEV is plug-in hybrid electric with diesel. In future, scenario.fuel_type should be parsed to see if it is a electric/gasoline PHEV, for example.
        assert (
            "cd_electricity" in fuels[0]
        ), r'fuels for electric/diesel PHEVs must be of format: ["cd_electricity", "cd_diesel", "cs_diesel"]'
        assert (
            "cd_diesel" in fuels[1]
        ), r'fuels for electric/diesel PHEVs must be of format: ["cd_electricity", "cd_diesel", "cs_diesel"]'
        assert (
            "cs_diesel" in fuels[2]
        ), r'fuels for electric/diesel PHEVs must be of format: ["cd_electricity", "cd_diesel", "cs_diesel"]'

    data = []
    for fuel_type, mpgge in zip(fuels, mpgges):
        data.append(
            [model_year, region, vehicle_segment_name, vocation, fuel_type, mpgge, age]
        )
    fefdata = pd.DataFrame(
        data,
        columns=[
            "Model Year",
            "Region",
            "Vehicle",
            "Vocation",
            "Fuel",
            "Fuel Efficiency [mi/gge]",
            "Age [yr]",
        ],
    )

    return fefdata


def fill_veh_expense_file(
    scenario: run_scenario.Scenario, cost_set: dict
) -> pd.DataFrame:
    """
    This helper method generates a dataframe of MSRP breakdown costs as Cost [$/veh]

    Args:
        scenario (run_scenario.Scenario): Scenario object of current selection
        cost_set (dict): Dictionary containing MSRP breakdown cost components

    Returns:
        vexpdf (pd.DataFrame): Dataframe containing MSRP components costs as Cost [$/veh]
    """
    # read in scenario values needed, such as segment name, vocation, etc.
    vehicle = scenario.segment_name
    vocation = scenario.vocation
    model_year = int(scenario.model_year)

    data = [
        [vehicle, vocation, model_year, cost_set["Glider"], "Glider"],
        [vehicle, vocation, model_year, cost_set["Fuel converter"], "Fuel converter"],
        [vehicle, vocation, model_year, cost_set["Fuel Storage"], "Fuel Storage"],
        [
            vehicle,
            vocation,
            model_year,
            cost_set["Motor & power electronics"],
            "Motor & power electronics",
        ],
        [vehicle, vocation, model_year, cost_set["Plug"], "Plug"],
        [vehicle, vocation, model_year, cost_set["Battery"], "Battery"],
        [
            vehicle,
            vocation,
            model_year,
            cost_set["Battery replacement"],
            "Battery replacement",
        ],
        [vehicle, vocation, model_year, cost_set["Purchase tax"], "Purchase tax"],
        # [vehicle, vocation, model_year, cost_set["Insurance"], "Insurance"],
        # [vehicle, vocation, model_year, cost_set["Residual Cost"], "Residual Cost"],
    ]
    vexpdf = pd.DataFrame(
        data, columns=["Vehicle", "Vocation", "Model Year", "Cost [$/veh]", "Category"]
    )

    return vexpdf


def fill_trav_exp_tsv(
    vehicle: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario
) -> pd.DataFrame:
    """
    This helper method generates a dataframe containing maintenance costs in Cost [$/mi]

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of analysis vehicle
        scenario (run_scenario.Scenario): Scenario object of current selection

    Returns:
        df (pd.DataFrame): Dataframe containing maintenance costs in Cost [$/mi]
    """

    columns = ["Year", "Region", "Vocation", "Vehicle", "Category", "Cost [$/mi]"]
    region = scenario.region
    vocation = scenario.vocation
    vehicle_name = scenario.segment_name
    maint = scenario.maint_oper_cost_dol_per_mi
    maint = list(np.float_(maint.strip(" ][").split(",")))
    veh_life_years = int(scenario.vehicle_life_yr)

    assert len(maint) == veh_life_years, (
        f"vehicle_life_yr of {veh_life_years} & length of input maint_oper_cost_dol_per_mi {len(maint)} do not align; "
        f"vehicle_life_yr life years & number of years in vmt should match\n"
        f"[vehicle_life_yr/[maint_oper_cost_dol_per_mi_1,...,maint_oper_cost_dol_per_mi_N]]:[{veh_life_years}/{maint}]"
    )
    data = []
    for i, yr in enumerate(
        range(int(scenario.model_year), int(scenario.model_year + veh_life_years))
    ):
        l1 = [yr, region, vocation, vehicle_name, "maintenance", maint[i]]
        data.append(l1)
        # data.append([yr, region, vocation, vehicle,  "payload opp cost", np.nan])
        # data.append([yr, region, vocation, vehicle,  "time opp cost",    np.nan])
        # data.append([yr, region, vocation, vehicle,  "labor opp cost",   np.nan])

    df = pd.DataFrame(data=data, columns=columns)

    return df


def fill_downtimelabor_cost_tsv(
    scenario: run_scenario.Scenario, oppy_cost_set: dict
) -> pd.DataFrame:
    """
    This helper method generates a dataframe containing fueling downtime and M&R downtime costs in Cost [$/Yr]

    Args:
        scenario (run_scenario.Scenario): Scenario object of current selection
        oppy_cost_set (dict): Dictionary containing fueling_downtime_oppy_cost_dol_per_yr,fueling_dwell_labor_cost_dol_per_yr and mr_downtime_oppy_cost_dol_per_yr

    Returns:
        df (pd.DataFrame): Dataframe containing fueling and MR downtime costs in Cost [$/Yr]
    """
    columns = ["Age [yr]", "Category", "Cost [$/Yr]"]
    data = []

    veh_life_years = int(scenario.vehicle_life_yr)

    fueling_downtime_oppy_cost_dol_per_yr = oppy_cost_set["fueling_downtime_oppy_cost_dol_per_yr"]
    fueling_dwell_labor_cost_dol_per_yr = oppy_cost_set["fueling_dwell_labor_cost_dol_per_yr"]
    mr_downtime_oppy_cost_dol_per_yr = oppy_cost_set["mr_downtime_oppy_cost_dol_per_yr"]
    # assert len(insurance_rates) >= veh_life_years, (f"vehLifeYears of {veh_life_years} & length of input insurance rates {len(insurance_rates)} do not align; "
    # f"vehLifeYears life years & number of years in vmt should match\n"
    # f"[vehLifeYears/[VMT_1,...,VMT_N]]:[{veh_life_years}/{insurance_rates}]")
    # \
    for i in range(0, veh_life_years):
        # downtime_costs_Dol = fueling_downtime_oppy_cost_dol[i] + MR_downtime_cost_Dol[i]
        # i is age, give it a VMT value for each entry in VMT or defer to last VMT entry
        data.append([i, "fueling downtime cost", fueling_downtime_oppy_cost_dol_per_yr[i]])
        data.append([i, "fueling labor cost", fueling_dwell_labor_cost_dol_per_yr[i]])
        data.append([i, "MR downtime cost", mr_downtime_oppy_cost_dol_per_yr[i]])

    df = pd.DataFrame(data, columns=columns)

    # print(f'insurance: {df}')
    return df


def fill_market_share_tsv(
    scenario: run_scenario.Scenario, num_vs: int = 1
) -> pd.DataFrame:
    """
    This helper method generates a dataframe containing market share of current vehicle selection per vehicle sold

    Args:
        scenario (run_scenario.Scenario): Scenario object of current selection
        num_vs (int, optional): Number of vehicles. Defaults to 1.

    Returns:
        df (pd.DataFrame): Dataframe containing market share of current vehicle in Market Share [veh/veh]
    """

    vocation = scenario.vocation
    vehicle = scenario.segment_name
    reg = scenario.region

    veh_life_years = int(scenario.vehicle_life_yr)
    model_year = int(scenario.model_year)
    # maint = list(np.float_(scenario.maint_oper_cost_dol_per_mi.strip('][').split(',')))
    data = []
    columns = ["Vehicle", "Vocation", "Model Year", "Region", "Market Share [veh/veh]"]
    data.append([vehicle, vocation, model_year, reg, 1 / num_vs])
    for yr in range(model_year + 1, model_year + veh_life_years):
        data.append([vehicle, vocation, yr, reg, 0])

    df = pd.DataFrame(data, columns=columns)

    return df


def fill_fuel_expense_tsv(
    vehicle: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario
) -> pd.DataFrame:
    """
    This helper method generates a dataframe of fuel operating costs in Cost [$/gge]

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of analysis vehicle
        scenario (run_scenario.Scenario): Scenario object of current selection

    Raises:
        Exception: Invalid fuel_type type

    Returns:
        df (pd.DataFrame): Dataframe containing fuel operating costs in Cost [$/gge]
    """

    fuels = scenario.fuel_type

    if vehicle.veh_pt_type in [gl.CONV, gl.HEV]:
        assert scenario.fuel_type[0].lower() in [
            "cng",
            "gasoline",
            "diesel",
            "hydrogen",
        ]
    elif vehicle.veh_pt_type in [gl.BEV]:
        assert scenario.fuel_type[0].lower() in ["electricity"]
    elif vehicle.veh_pt_type == gl.PHEV:
        assert (
            "cd_electricity" in fuels[0]
        ), r'fuels must be of format: ["cd_electricity", "cd_diesel", "cs_diesel"]'
        assert (
            "cd_diesel" in fuels[1]
        ), r'fuels must be of format: ["cd_electricity", "cd_diesel", "cs_diesel"]'
        assert (
            "cs_diesel" in fuels[2]
        ), r'fuels must be of format: ["cd_electricity", "cd_diesel", "cs_diesel"]'

    veh_life_span = int(scenario.vehicle_life_yr)
    cat = "Fuel"
    columns = ["Year", "Fuel", "Category", "Cost [$/gge]"]
    data = []
    regdf = pd.read_csv(gl.REGIONAL_FUEL_PRICES_BY_TYPE_BY_YEAR)
    regdf = regdf.set_index("Fuel")
    for fuel_type in fuels:
        # cat = fuel_type
        for yr in range(
            int(scenario.model_year), int(scenario.model_year + veh_life_span)
        ):
            regdf = regdf[regdf["Region"] == scenario.region]
            # all costs are converted to $ per gallon gasoline equivalent
            # TODO, may want to be more explicit than just finding substrings
            if "diesel" in fuel_type.lower() and "bio" not in fuel_type.lower():
                dieselDolPerGal = regdf.loc["dieselDolPerGal", str(yr)]
                Dslgge = 1 * (33.7 / 37.95)
                cost = dieselDolPerGal * Dslgge
            elif "gasoline" in fuel_type.lower():
                gasolineDolPerGal = regdf.loc["gasolineDolPerGal", str(yr)]
                cost = gasolineDolPerGal
            elif "electricity" in fuel_type.lower():
                dolPerKwh = regdf.loc["dolPerKwh", str(yr)]
                cost = dolPerKwh * 33.7  # 33.41 kwh per gallon of gasoline
            elif fuel_type.lower() == "cng":
                CNGDolPerGge = regdf.loc["CNGDolPerGge", str(yr)]
                cost = CNGDolPerGge
            elif fuel_type.lower() == "hydrogen":
                hydrogenDolPerGGE = regdf.loc["hydrogenDolPerGGE", str(yr)]
                cost = hydrogenDolPerGGE
            else:
                raise Exception(
                    f"TCO fuel calc: fill_fuel_expense_tsv:: unknown fuel type {fuel_type}"
                )
            data.append([yr, fuel_type, cat, cost])

            # scenario.scenario_gge_regional_temporal_fuel_price += f"fuel: {fuel_type}; region: {scenario.region}; year: {yr}; $_gge: {cost}"

    df = pd.DataFrame(data, columns=columns)

    return df


def fill_annual_tsv(scenario: run_scenario.Scenario) -> pd.DataFrame:
    """
    This helper method generates a dataframe of annual vehicle miles traveled (vmt) - Annual Travel [mi/yr]

    Args:
        scenario (run_scenario.Scenario): Scenario object of current selection

    Returns:
        df (pd.DataFrame): Dataframe containing Annual Travel [mi/yr]
    """
    columns = ["Age [yr]", "Annual Travel [mi/yr]"]
    data = []

    veh_life_years = int(scenario.vehicle_life_yr)

    vmt = scenario.vmt
    assert len(vmt) == veh_life_years, (
        f"vehicle_life_yr of {veh_life_years} & length of input vmt {len(vmt)} do not align; "
        f"vehicle_life_yr life years & number of years in vmt should match\n"
        f"[vehicle_life_yr/[VMT_1,...,VMT_N]]:[{veh_life_years}/{vmt}]"
    )

    for i in range(0, veh_life_years):
        miles = vmt[i]
        # i is age, give it a vmt value for each entry in vmt or defer to last vmt entry
        data.append([i, miles])

    df = pd.DataFrame(data, columns=columns)

    return df


def fill_reg_sales_tsv(
    scenario: run_scenario.Scenario, num_vs: int = 1
) -> pd.DataFrame:
    """
    This helper method generates a dataframe containing vehicle sales per year - Sales [veh]

    Args:
        scenario (run_scenario.Scenario): Scenario object of current selection
        num_vs (int, optional): Number of vehicles. Defaults to 1.

    Returns:
        df (pd.DataFrame): Dataframe containing vehicle sales in Sales [veh]
    """
    veh_life_years = int(scenario.vehicle_life_yr)
    model_year = int(scenario.model_year)
    reg = scenario.region

    columns = ["Model Year", "Region", "Sales [veh]"]
    data = []
    for yr in range(model_year, veh_life_years + model_year):
        if yr == model_year:
            data.append([yr, reg, num_vs])
        else:
            data.append([yr, reg, 0])

    df = pd.DataFrame(data, columns=columns)

    return df


def fill_insurance_tsv(
    scenario: run_scenario.Scenario, veh_cost_set: dict
) -> pd.DataFrame:
    """
    This helper method generates a dataframe containing vehicle insurance costs as Cost [$/Yr]

    Args:
        scenario (run_scenario.Scenario): Scenario object of current selection
        veh_cost_set (dict): Dictionary containing MSRP costs

    Returns:
        df (pd.DataFrame): Dataframe containing insurance costs in Cost [$/Yr]
    """
    columns = ["Age [yr]", "Category", "Cost [$/Yr]"]
    data = []

    veh_life_years = int(scenario.vehicle_life_yr)

    insurance_rates = scenario.insurance_rates_pct_per_yr
    assert len(insurance_rates) >= veh_life_years, (
        f"vehicle_life_yr of {veh_life_years} & length of input insurance rates {len(insurance_rates)} do not align; "
        f"vehicle_life_yr life years & number of years in vmt should match\n"
        f"[vehicle_life_yr/[VMT_1,...,VMT_N]]:[{veh_life_years}/{insurance_rates}]"
    )

    MSRP = veh_cost_set["msrp"]
    for i in range(0, veh_life_years):
        insurance_costperyear = insurance_rates[i] * MSRP / 100
        # i is age, give it a vmt value for each entry in vmt or defer to last vmt entry
        data.append([i, "insurance", insurance_costperyear])

    df = pd.DataFrame(data, columns=columns)

    # print(f'insurance: {df}')
    return df


def fill_residual_cost_tsc(
    vehicle: fastsim.vehicle.Vehicle,
    scenario: run_scenario.Scenario,
    veh_cost_set: dict,
) -> pd.DataFrame:
    """
    This helper method generates a dataframe of residual costs as Cost [$/Yr]

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of analysis vehicle
        scenario (run_scenario.Scenario): Scenario object of current selection
        veh_cost_set (dict): Dictionary containing MSRP costs

    Returns:
        df (pd.DataFrame): Dataframe containing vehicle residual costs as Cost [$/Yr]
    """
    columns = ["Age [yr]", "Category", "Cost [$/Yr]"]
    data = []

    veh_life_years = int(scenario.vehicle_life_yr)

    residual_rate = find_residual_rates(vehicle, scenario)

    # assert len(insurance_rates) >= veh_life_years, (f"vehicle_life_yr of {veh_life_years} & length of input insurance rates {len(insurance_rates)} do not align; "
    # f"vehicle_life_yr life years & number of years in vmt should match\n"
    # f"[vehicle_life_yr/[VMT_1,...,VMT_N]]:[{veh_life_years}/{insurance_rates}]")

    MSRP = veh_cost_set["msrp"]
    for i in range(0, veh_life_years):
        if i == veh_life_years - 1:
            residual_cost = -residual_rate * MSRP
        else:
            residual_cost = 0
        data.append([i, "residual cost", residual_cost])

    df = pd.DataFrame(data, columns=columns)

    # print(f'residual: {df}')
    return df


def fill_survival_tsv(scenario: run_scenario.Scenario, num_vs=1) -> pd.DataFrame:
    """
    This helper method generates a dataframe containing surviving vehicles as Surviving Vehicles [veh/veh]

    Args:
        scenario (run_scenario.Scenario): Scenario object of current selection
        num_vs (int, optional): Number of vehicles. Defaults to 1.

    Returns:
        df (pd.DataFrame): Dataframe containing number of surviving vehicles on road in Surviving Vehicles [veh/veh]
    """

    columns = ["Age [yr]", "Surviving Vehicles [veh/veh]"]
    data = []
    veh_life_span = int(scenario.vehicle_life_yr)
    for yr in range(0, veh_life_span):
        data.append([yr, num_vs / num_vs])

    df = pd.DataFrame(data, columns=columns)

    return df


def fill_fuel_split_tsv(
    vehicle: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario, mpgge: dict
) -> pd.DataFrame:
    """
    This helper method generates a dataframe of fraction of travel in each fuel type as Fraction of Travel [mi/mi]

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of analysis vehicle
        scenario (run_scenario.Scenario): Scenario object of current selection
        mpgge (dict): MPGGE dictionary from fueleconomy.get_mpgge()

    Returns:
        df (pd.DataFrame): Dataframe containing fraction of travel in each fuel type as Fraction of Travel [mi/mi]
    """
    columns = ["Vehicle", "Fuel", "Vocation", "Fraction of Travel [mi/mi]"]
    vocation = scenario.vocation
    vehicle_segment = scenario.segment_name
    fuels = scenario.fuel_type

    if vehicle.veh_pt_type == gl.PHEV:
        uf = run_scenario.get_phev_util_factor(scenario, vehicle, mpgge)

    if vehicle.veh_pt_type != gl.PHEV:
        frac_time_traveled_mi_mi = 1
        data = [[vehicle_segment, fuels[0], vocation, frac_time_traveled_mi_mi]]
    else:
        data = [
            [vehicle_segment, fuels[0], vocation, uf],  # cd_electricity
            [vehicle_segment, fuels[1], vocation, uf],  # cd_diesel
            [vehicle_segment, fuels[2], vocation, 1 - uf],  # cs_diesel
        ]

    df = pd.DataFrame(data, columns=columns)

    return df



import copy
import logging
import time
from typing import Tuple
import warnings
from time import gmtime, strftime

import autograd.numpy as anp
import fastsim
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# from pymoo.algorithms.nsga2 import NSGA2 as NSGA2
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.algorithms.soo.nonconvex.nelder import NelderMead
from pymoo.algorithms.soo.nonconvex.pattern import PatternSearch

# from pymoo.algorithms.so_local_search import LocalSearch
from pymoo.algorithms.soo.nonconvex.pso import PSO
import pymoo.core
from pymoo.core.problem import ElementwiseProblem
import pymoo.core.result
from pymoo.operators.sampling.lhs import LatinHypercubeSampling as LHS

# pymoo stuff
from pymoo.optimize import minimize
from pymoo.termination.default import DefaultMultiObjectiveTermination as MODT
from pymoo.termination.ftol import MultiObjectiveSpaceTermination
from pymoo.util.display.output import Output
import pymoo
from t3co.run import Global as gl
from t3co.run import run_scenario

# PyMoo runs a vehicle optimization with POC accounted for that produces 3 designs that
# meet accel and grade targets and are within 1% of target range.  Grant says this is
# about 2x as fast as what the other version did for a single vehicle.  We were working on
# - speed up (got it about 35x faster with ~2 person-hours of work)
# - validation -- check
# - making it callable as python module

# optimization objectives
TCO = "tot_cost"
PHEV_MINIMIZE_FUEL_USE_OBJECTIVE = "objective_phev_minimize_fuel_use"
OBJECTIVES = [TCO, PHEV_MINIMIZE_FUEL_USE_OBJECTIVE]

# optimization constraints
(
    RANGE,
    ACCEL30,
    ACCEL60,
    GRADE125,
    GRADE6,
) = (
    "range_mi",
    "zero_to_30_loaded",
    "zero_to_60_loaded",
    "grade_1_25_mph_ach",
    "grade_6_mph_ach",
)
ACCEL_GRADE_OVERSHOOT = "accel_grade_overshoot_tol_constraint"
TRACE_MISS_DIST_PERCENT = "constraint_trace_miss_dist_percent_on"
C_RATE = "c_rate"
PHEV_MINIMIZE_FUEL_USE_CONSTRAINT = "PHEV_MINIMIZE_CD_FUEL_USE_PERCENTAGE_CONSTRAINT"
CONSTRAINTS = [
    RANGE,
    ACCEL30,
    ACCEL60,
    GRADE125,
    GRADE6,
    ACCEL_GRADE_OVERSHOOT,
    TRACE_MISS_DIST_PERCENT,
    C_RATE,
    PHEV_MINIMIZE_FUEL_USE_CONSTRAINT,
]


# optimization parameters
# scalar powertrain modifers
KNOB_FCMAXKW = "fcMaxOutKw"
KNOB_ess_max_kwh = "ess_max_kwh"
KNOB_mc_max_kw = "mc_max_kw"
KNOB_fs_kwh = "fs_kwh"
# COST:  b * cda_perc^2 + a * cda_perc
# MASS:  b * cda_perc^2 + a * cda_perc
#   (need COST {a,b} and MASS {a,b})
KNOB_CDA = "CdA_perc_imp"
# curves
KNOB_FCPEAKEFF = "fc_peak_eff"
KNOB_WTDELTAPERC = "wt_delta_perc"

# also new variables to refer to light-weighting curve data,
#   fc peak eff curve data, & CdA (need COST {a,b} and MASS {a,b})
# new inputs for TDA/Benefits
# CdA_perc_imp_bounds | (5, 30)
# COST  b * cda_perc^2 + a * cda_perc
# MASS  b * cda_perc^2 + a * cda_perc

# T3CO moo return codes
OPTIMIZATION_SUCCEEDED = 1
OPTIMIZATION_FAILED_TO_CONVERGE = 2
EXCEPTION_THROWN = 3


ALGO_NSGA2 = "NSGA2"
ALGO_NelderMead = "NelderMead"
ALGO_PatternSearch = "PatternSearch"
ALGO_PSO = "PSO"

ALGORITHMS = [
    ALGO_NSGA2,
    ALGO_NelderMead,
    ALGO_PatternSearch,
    ALGO_PSO,
]

KNOBS = [
    KNOB_CDA,
    KNOB_FCMAXKW,
    KNOB_ess_max_kwh,
    KNOB_mc_max_kw,
    KNOB_FCPEAKEFF,
    KNOB_WTDELTAPERC,
    KNOB_fs_kwh,
]


class T3COProblem(ElementwiseProblem):
    """
    Class for creating PyMoo problem.
    """

    moobasevehicle: fastsim.vehicle
    mooadvancedvehicle: fastsim.vehicle
    opt_scenario: run_scenario.Scenario
    designcycle: fastsim.cycle
    config: run_scenario.Config

    def setup_opt_records(self):
        """
        This method sets up the empty optimization record arrays
        """
        #

        # objectives
        self.r_tcos = []
        self.r_cd_fc_kwh_percent = []
        self.r_cd_fc_kwh_used = []
        self.r_cd_elec_kwh_used = []

        self.reporting_vars = None
        self.r_grade_6s = []
        self.r_grade_1p25s = []
        self.r_accel_30l = []
        self.r_accel_60l = []
        self.r_ranges = []
        self.r_fuel_efficiencies = []
        self.r_wt_delta_perc_guess = []
        self.r_CdA_reduction_perc = []
        self.r_fc_peak_eff_guess = []
        self.r_fc_max_out_kw_guess = []
        self.r_fs_kwh_guess = []
        self.r_max_ess_kwh_guess = []
        self.r_max_motor_kw_guess = []
        # constraint records
        self.accel_30_constraint = []
        self.accel_60_constraint = []
        self.grade_6_constraint = []
        self.grade_1p25_constraint = []
        self.range_constraint = []
        self.grade_accel_overshoot_tol_constraint = []
        self.c_rate_constraint = []
        self.trace_miss_distance_percent_constraint_record = []
        self.phev_min_fuel_use_prcnt_const_record = []

    def __init__(
        self,
        knobs_bounds: dict,
        vnum: float,
        optimize_pt: str = gl.BEV,
        obj_list: list = None,
        constr_list: list = None,
        verbose: bool = False,
        config: run_scenario.Config = None,
        **kwargs,
    ) -> None:
        """
        This constructor initializes optimization input variables

        Args:
            knobs_bounds (dict): Dictionary containing knobs bounds for optimization
            vnum (float): Vehicle selection number
            optimize_pt (str, optional): Vehicle powertrain type - Conv, BEV, HEV, PHEV. Defaults to gl.BEV.
            obj_list (list, optional): List of objectives. Defaults to None.
            constr_list (list, optional): List of constraints. Defaults to None.
            verbose (bool, optional): if True, prints process steps. Defaults to False.
            config (run_scenario.Config, optional): T3CO Config object containing analysis attributes and scenario attribute overrides. Defaults to None.

        """
        self.setup_opt_records()

        # TODO there should probably not be any default values for kwargs

        # TODO: figure out parallelization and then modify the following line accordingly
        _ = kwargs.pop("parallelization", None)

        # possible TODO: make this a dict for grade, accel, and range tolerance
        self.range_overshoot_tol = kwargs.pop("range_overshoot_tol", None)
        self.grade_accel_overshoot_tol = kwargs.pop(
            "grade_accel_overshoot_tol", 0.01
        )  # TODO default of 1% # user should explicitly turn this on and set this outside moo.py

        # MOO optimization interpolation arrays and coefficients
        # engine peak efficiecy percents
        self.fc_eff_array = kwargs.pop(
            "fc_eff_array", np.array([46, 46, 51, 51, 53, 53, 60]) / 100
        )
        # additional $ cost per kw of engine per engine efficiency percent
        self.fc_cost_coeff_array = kwargs.pop(
            "fc_cost_coeff_array", np.array([0, 0, 7, 16, 22, 40, 74])
        )
        # glider_kg light-weighted percents
        self.ltwt_delta_percs = kwargs.pop(
            "ltwt_delta_percs", np.array([0, 5, 5, 10, 10, 25]) / 100
        )
        # additional $ cost per kg light-weighted
        self.ltwt_dol_per_kg_costs = kwargs.pop(
            "ltwt_dol_per_kg_costs", np.array([4.41, 8.82, 13.23, 17.64, 22.05, 35.27])
        )
        # , if you want to keep default values, they should be 52.1 and 6.7
        # To make it like what I sent before, Arthur can recalculate (should be enough decimals):
        # a = 100*a
        # b = 10000*b - Alicia
        self.cda_cost_coeff_a = kwargs.pop("cda_cost_coeff_a", 52.053)
        self.cda_cost_coeff_b = kwargs.pop("cda_cost_coeff_b", 6.7018)
        self.cda_mass_coeff_a = kwargs.pop(
            "cda_mass_coeff_a", 0
        )  # default formula doesn't need a_M coeff, wt = 17967.9*p^2
        self.cda_mass_coeff_b = kwargs.pop("cda_mass_coeff_b", 1.79679)
        self.cda_perc_imp_at_which_wt_penalty_maxes_out = kwargs.pop(
            "cda_perc_imp_at_which_wt_penalty_maxes_out", 9999999999
        )

        # just checker values, knobs_bounds may not actually have all of these populated, so [0,0] is never used beyond the asserts
        max_fc_bound = knobs_bounds.get(KNOB_FCPEAKEFF, (0, 0))[1]
        max_ltwt_bound = knobs_bounds.get(KNOB_WTDELTAPERC, (0, 0))[1]
        CdA_perc_imp_knob_max = knobs_bounds.get(KNOB_CDA, (0, 0))[1]

        self.knobs_bounds = knobs_bounds

        knobs = [key for key in knobs_bounds.keys()]
        lower_bounds = anp.array([val[0] for bound, val in knobs_bounds.items()])
        upper_bounds = anp.array([val[1] for bound, val in knobs_bounds.items()])

        self.write_tsv = kwargs.pop("write_tsv", False)

        self.obj_list = obj_list
        if obj_list is None:
            # create default objective list
            self.obj_list = [TCO]

        self.constr_list = constr_list

        # add acceleration and grade overshoot tolerance constraint implicitly if user is apply grade or acceleration constraints
        if len(set(self.constr_list) & set([ACCEL30, ACCEL60, GRADE125, GRADE6])):
            self.constr_list.append(ACCEL_GRADE_OVERSHOOT)

        n_constr = len(self.constr_list)
        # input (f"moo.py: T3COProblem __init__\n{self.constr_list}, \neffective constraints: {n_constr}")
        # make it so that knobs always need to be specified.
        self.knobs = knobs

        self.verbose = verbose

        self.optimize_pt = optimize_pt

        self.instantiate_moo_vehicles_and_scenario(vnum, config)

        # time dilation options, turned on for fuel efficiency cycle
        if "missed_trace_correction" in kwargs:
            self.opt_scenario.missed_trace_correction = kwargs.pop(
                "missed_trace_correction"
            )
            self.opt_scenario.max_time_dilation = kwargs.pop("max_time_dilation")
            self.opt_scenario.min_time_dilation = kwargs.pop("min_time_dilation")
            self.opt_scenario.time_dilation_tol = kwargs.pop("time_dilation_tol")

        n_obj = len(self.obj_list)

        n_args = len(knobs)

        # asserts
        assert (
            self.optimize_pt == self.mooadvancedvehicle.veh_pt_type
        ), "instantiate_moo_vehicles_and_scenario should have handled this"
        for knob in self.knobs:
            assert knob in KNOBS, f"knob {knob} not in defined parameters: {KNOBS}"
        for constraint in self.constr_list:
            assert (
                constraint in CONSTRAINTS
            ), f"constraint {constraint} not in defined constraints: {CONSTRAINTS}"
            if constraint == PHEV_MINIMIZE_FUEL_USE_CONSTRAINT:
                assert (
                    self.opt_scenario.constraint_phev_minimize_fuel_use_percent > 0
                    and self.opt_scenario.constraint_phev_minimize_fuel_use_percent < 1
                )
        assert len(OBJECTIVES) >= 1, "enable at least one objective"
        for obj in self.obj_list:
            assert (
                obj in OBJECTIVES
            ), f"objective {obj} not in defined objectives: {OBJECTIVES}"
        assert (
            set(self.constr_list) & set(self.obj_list) == set()
        ), f"cannot have constraints and objectives overlap constraints {self.constr_list} & {self.obj_list}"
        if KNOB_fs_kwh in self.knobs:
            assert (
                self.optimize_pt in [gl.CONV, gl.HEV, gl.PHEV]
            ), f"input knob bounds: {knobs_bounds}\naltering fuel store (tank, not battery) kwh size but vehicle powertrain to be optimized is {self.optimize_pt}"
        if KNOB_FCMAXKW in self.knobs:
            assert (
                self.optimize_pt in [gl.CONV, gl.HEV, gl.PHEV]
            ), f"input knob bounds: {knobs_bounds}\naltering fuel converter size but vehicle powertrain is {self.optimize_pt}"
        if KNOB_ess_max_kwh in self.knobs:
            assert (
                self.optimize_pt in [gl.BEV, gl.HEV, gl.PHEV]
            ), f"input knob bounds: {knobs_bounds}\naltering battery size but vehicle powertrain to be optimized is {self.optimize_pt}"
        if KNOB_mc_max_kw in self.knobs:
            assert (
                self.optimize_pt in [gl.BEV, gl.HEV, gl.PHEV]
            ), f"input knob bounds: {knobs_bounds}\naltering motor size but vehicle powertrain to be optimized is {self.optimize_pt}"
        assert (
            max_fc_bound <= max(self.fc_eff_array)
        ), f"max eng eff knob val greater than max of eng eff array {max_fc_bound} > {self.fc_eff_array}, could yield free efficiency"
        assert (
            max_ltwt_bound <= max(self.ltwt_delta_percs)
        ), f"max light-weight knob val greater than max of light-weighting percent array {max_ltwt_bound} > {self.ltwt_delta_percs}, could yield free light-weighting"
        assert (
            max_fc_bound <= 1
        ), f"knob bounds must be decimal format, max_fc_bound: {max_fc_bound}"
        assert (
            max_ltwt_bound <= 1
        ), f"knob bounds must be decimal format, max_ltwt_bound: {max_ltwt_bound}"
        assert (
            CdA_perc_imp_knob_max <= 1
        ), f"knob bounds must be decimal format, CdA_perc_imp_knob_max: {CdA_perc_imp_knob_max}"

        # n_ieq_constr, number of constraints that must yield < 0
        super().__init__(
            n_var=n_args,
            n_obj=n_obj,
            n_ieq_constr=n_constr,
            # lower bounds
            xl=lower_bounds,
            # upper bounds
            xu=upper_bounds,
            **kwargs,
        )

        if len(kwargs) > 0:
            warnings.warn(
                f"Possible unused/invalid kwargs provided:\n {list(kwargs.keys())}"
            )

    def compile_reporting_vars(self) -> None:
        """
        This method creates an output dictionary containing optimization results
        """
        d = {
            "objective_TCOs": self.r_tcos,
            "objective_fc_khw_percent": self.r_cd_fc_kwh_percent,
            "objective_fc_khw_used": self.r_cd_fc_kwh_used,
            "objective_elec_khw_used": self.r_cd_elec_kwh_used,
            "r_mph_ach_grade_6s": self.r_grade_6s,
            "r_mph_ach_grade_1p25s": self.r_grade_1p25s,
            "r_sec_to_ach_30mph_ldd": self.r_accel_30l,
            "r_sec_to_ach_60mph_ldd": self.r_accel_60l,
            "r_ach_ranges_mi": self.r_ranges,
            "target_mph_grade_6s": [
                self.opt_scenario.min_speed_at_6pct_grade_in_5min_mph
            ]
            * len(self.r_grade_6s),
            "target_mph_grade_1p25s": [
                self.opt_scenario.min_speed_at_1p25pct_grade_in_5min_mph
            ]
            * len(self.r_grade_6s),
            "target_sec_to_30mph_ldd": [self.opt_scenario.max_time_0_to_30mph_at_gvwr_s]
            * len(self.r_grade_6s),
            "target_sec_to_60mph_ldd": [self.opt_scenario.max_time_0_to_60mph_at_gvwr_s]
            * len(self.r_grade_6s),
            "target_range_mi": [self.opt_scenario.target_range_mi]
            * len(self.r_grade_6s),
            "r_ach_fuel_efficiencies": self.r_fuel_efficiencies,
            "accel_30_constraint_vals": self.accel_30_constraint,
            "accel_60_constraint_vals": self.accel_60_constraint,
            "grade_6_constraint_vals": self.grade_6_constraint,
            "grade_1p25_constraint_vals": self.grade_1p25_constraint,
            "range_constraint_vals": self.range_constraint,
            "grade_accel_overshoot_tol_constraint_vals": self.grade_accel_overshoot_tol_constraint,
            "trace_miss_dist_percent_constraint_vals": self.trace_miss_distance_percent_constraint_record,
            "phev_min_fuel_use_pct_constraint_vals": self.phev_min_fuel_use_prcnt_const_record,
            "r_wt_delta_perc_guess": self.r_wt_delta_perc_guess,
            "r_CdA_reduction_perc": self.r_CdA_reduction_perc,
            "r_fc_peak_eff_guess": self.r_fc_peak_eff_guess,
            "r_fc_max_out_kw_guess": self.r_fc_max_out_kw_guess,
            "r_fs_kwh_guess": self.r_fs_kwh_guess,
            "r_max_ess_kwh_guess": self.r_max_ess_kwh_guess,
            "r_max_motor_kw_guess": self.r_max_motor_kw_guess,
        }
        self.reporting_vars = pd.DataFrame(data=d)

    def instantiate_moo_vehicles_and_scenario(self, vnum: int, config=None) -> None:
        """
        This method instantiates the multi-objective optimization problem vehicles and scenarios, starting with the baseline Conventional vehicle.

        Args:
            vnum (int): vehicle selection number
            config (run_scenario.Config, optional): T3CO Config object containing analysis attributes and scenario attribute overrides. Defaults to None.

        Raises:
            TypeError: Invalid optimize_pt selection
        """
        self.moobasevehicle = run_scenario.get_vehicle(
            vnum,
            veh_input_path=gl.FASTSIM_INPUTS,
        )

        self.opt_scenario, self.designcycle = run_scenario.get_scenario_and_cycle(
            vnum, gl.OTHER_INPUTS, config=config
        )

        if (
            self.opt_scenario.fuel_type == "diesel and electricity"
            and self.moobasevehicle.veh_pt_type == gl.HEV
        ):
            self.opt_scenario.fuel_type = "diesel"

        # save baseline values for optimization diffs
        self.opt_scenario.originalGliderPrice = (
            self.opt_scenario.vehicle_glider_cost_dol
        )
        self.opt_scenario.originalglider_kg = self.moobasevehicle.glider_kg
        self.opt_scenario.originalIceDolPerKw = self.opt_scenario.fc_ice_cost_dol_per_kw
        self.opt_scenario.origfc_eff_map = self.moobasevehicle.fc_eff_map.copy()
        self.opt_scenario.originalcargo_kg = self.moobasevehicle.cargo_kg
        self.opt_scenario.originaldrag_coef = self.moobasevehicle.drag_coef

        self.mooadvancedvehicle = copy.copy(self.moobasevehicle)

        if self.optimize_pt == gl.BEV:
            self.mooadvancedvehicle.veh_pt_type = gl.BEV
            run_scenario.set_max_fuel_converter_kw(self.mooadvancedvehicle, 0)
            # change to 0, based on Excel version direction when trying to run as EV
            self.mooadvancedvehicle.fs_max_kw = 0
            run_scenario.set_fuel_store_kwh(self.mooadvancedvehicle, 0)
        elif self.optimize_pt == gl.CONV:
            self.mooadvancedvehicle.veh_pt_type = gl.CONV
            run_scenario.set_max_motor_kw(
                self.mooadvancedvehicle, scenario=None, max_motor_kw=0
            )
            run_scenario.set_max_battery_kwh(self.mooadvancedvehicle, max_ess_kwh=0)
        elif self.optimize_pt == gl.HEV:
            self.mooadvancedvehicle.veh_pt_type = gl.HEV
            # nothing to zero out for HEV/FCEV
        elif self.optimize_pt == gl.PHEV:
            self.mooadvancedvehicle.veh_pt_type = gl.PHEV
        else:
            raise TypeError(f"invalid optimize_pt selection {self.optimize_pt}")

        # establish payload opp cost calc module
        # TODO, this should include FCEV, and this TCO element should be True/False activated
        # if self.opt_scenario.activate_tco_payload_cap_cost_multiplier or self.opt_scenario.activate_tco_fueling_dwell_time_cost or self.opt_scenario.activate_mr_downtime_cost:
        #     self.oppcostobj = opportunity_cost.OpportunityCost(self.opt_scenario)

    # --------- optimizer parameter application methods ---------
    def cda_percent_delta_knob(
        self, CdA_perc_reduction: str, optvehicle: fastsim.vehicle.Vehicle
    ) -> None:
        """
        This method sets the drag_coef based on aero improvement curve and glider_kg based on cda_cost_coeff_a and cda_cost_coeff_b

        Args:
            CdA_perc_reduction (str): Name of aero improvement curve file
            optvehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object for optimization vehicle

        """

        # check value of CdA_reduction
        # print(CdA_perc_reduction)

        # functions that return cost, weight, and possibly other penalties w.r.t. various improvements

        def get_cost_per_CdA_delta(x):
            """
            Returns cost penalty given percent improvement in CdA.
            """
            # COST:  a * cda_perc + b * cda_perc^2
            a_C = self.cda_cost_coeff_a
            b_C = self.cda_cost_coeff_b
            x = x * 100
            return a_C * x + b_C * x**2

        def get_mass_per_CdA_delta(x):
            """
            Returns weight penalty [Lb converted to kg] given % improvement in CdA.
            These formulas are developed in pounds. Converted to kg for FASTSim at the end.
            """
            # MASS: a * cda_perc + b * cda_perc^2
            a_M = (
                self.cda_mass_coeff_a
            )  # default formula doesn't need a_M coeff, wt = 17967.9*p^2
            b_M = self.cda_mass_coeff_b
            x = x * 100
            CdA_perc_imp_at_which_wt_penalty_maxes_out = (
                self.cda_perc_imp_at_which_wt_penalty_maxes_out
            )
            return min(
                gl.lbs_to_kgs(
                    a_M * CdA_perc_imp_at_which_wt_penalty_maxes_out
                    + b_M * CdA_perc_imp_at_which_wt_penalty_maxes_out**2
                ),
                gl.lbs_to_kgs(a_M * x + b_M * x**2),
            )

        # reset drag coefficient based on percent improvement
        optvehicle.drag_coef = max(
            0.01,  # this is an additional hard-coded bound on drag coefficient that should probably be removed (throw error or warning if drag_coef < 0.01 instead of running and replacing with this default value)
            self.opt_scenario.originaldrag_coef * (1 - CdA_perc_reduction),
        )
        # glider cost penalty due to CdA improvement
        CdA_cost = get_cost_per_CdA_delta(CdA_perc_reduction)
        self.opt_scenario.vehicle_glider_cost_dol = (
            self.opt_scenario.vehicle_glider_cost_dol + CdA_cost
        )
        # mass adjustments. Add CdA mass FIRST, if applicable. Then limit cargo_kg if overweight from battery.
        optvehicle.glider_kg = optvehicle.glider_kg + get_mass_per_CdA_delta(
            CdA_perc_reduction
        )
        optvehicle.set_veh_mass()

    def weight_delta_percent_knob(
        self, wt_perc_reduction: float, optvehicle: fastsim.vehicle.Vehicle
    ) -> None:
        """
        This method sets the knob from the lightweighting curve

        Args:
            wt_perc_reduction (float): Weight reduction percentage value from lightweighting curve
            optvehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of the optimization vehicle
        """
        wt_delta_cost_per_kg = np.interp(
            x=wt_perc_reduction, xp=self.ltwt_delta_percs, fp=self.ltwt_dol_per_kg_costs
        )
        wt_delta_kg = optvehicle.glider_kg * wt_perc_reduction
        # Calculate lightweight cost from curve as an integral
        x_new = (
            self.ltwt_delta_percs[self.ltwt_delta_percs <= wt_perc_reduction]
            * optvehicle.glider_kg
        )
        y_new = self.ltwt_dol_per_kg_costs[self.ltwt_delta_percs <= wt_perc_reduction]
        if wt_delta_kg not in x_new:
            x_new = np.append(x_new, wt_delta_kg)
            y_new = np.append(y_new, wt_delta_cost_per_kg)
        self.opt_scenario.vehicle_glider_cost_dol = (
            self.opt_scenario.vehicle_glider_cost_dol + np.trapz(y_new, x_new)
        )
        optvehicle.glider_kg = optvehicle.glider_kg - wt_delta_kg
        optvehicle.set_veh_mass()

    def fc_peak_eff_knob(
        self, fc_peak_eff: float, optvehicle: fastsim.vehicle.Vehicle
    ) -> None:
        """
        This method sets the knob from the engine efficiency curve

        Args:
            fc_peak_eff (float): Fuel converter peak effiency override from engine efficiency improvement curve
            optvehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object for optimization vehicle
        """
        fc_eff_array = self.fc_eff_array
        cost_coeff_array = self.fc_cost_coeff_array
        d_eff_dol_per_kw = np.interp(fc_peak_eff, fc_eff_array, cost_coeff_array)
        self.opt_scenario.fc_ice_cost_dol_per_kw = (
            self.opt_scenario.originalIceDolPerKw + d_eff_dol_per_kw
        )
        self.adjust_fc_peak_eff(fc_peak_eff, self.opt_scenario, optvehicle)

    def get_objs(
        self, x: dict, write_tsv: bool = False
    ) -> Tuple[np.array, np.array, dict]:
        """
        This method gets called when PyMoo calls _evaluate. It initializes objectives and constraints and runs vehicle_scenario_sweep

        x optimization knobs = [max motor kw, battery kwh, drag coeff % improvement]
        Function for running FE cycles and accel tests then returning
        fuel consumption and zero-to-sixty times.

        x is a set of genes (or parameters), so kwh size is a gene
        chromosome is a full gene, all values in x

        Args:
            x (dict): Dictionary containing optimization knobs - {max motor kw, battery kwh, drag coeff % improvement}
            write_tsv (bool, optional): if True, save intermediate dataframes. Defaults to False.

        Returns:
            obj_arr_F (np.array): Array of objectives - tot_cost and phev_cd_fuel_used_kwh
            constraint_results_G (np.array): Array of constraints
            rs_sweep (dict): Output dictionary from vehicle_scenario_sweep
        """

        # dict for providing mechanism for making sure all knobs get used
        x_dict = {knob: x[self.knobs.index(knob)] for knob in self.knobs}

        # if self.optimize_pt not in [gl.CONV, gl.BEV, gl.HEV]:
        #     raise TypeError(f"optimize_pt is not configured for {self.optimize_pt}")
        designcycle = self.designcycle
        optvehicle = self.mooadvancedvehicle

        # reset glider price and weight for light-weighting and/or CdA percent improvement
        self.opt_scenario.vehicle_glider_cost_dol = (
            self.opt_scenario.originalGliderPrice
        )
        optvehicle.glider_kg = self.opt_scenario.originalglider_kg

        wt_delta_perc_guess = x_dict.pop(KNOB_WTDELTAPERC, None)
        CdA_reduction_perc = x_dict.pop(KNOB_CDA, None)
        fc_peak_eff_guess = x_dict.pop(KNOB_FCPEAKEFF, None)
        fc_max_out_kw_guess = x_dict.pop(KNOB_FCMAXKW, None)
        max_ess_kwh_guess = x_dict.pop(KNOB_ess_max_kwh, None)
        max_motor_kw_guess = x_dict.pop(KNOB_mc_max_kw, None)
        fs_kwh_guess = x_dict.pop(KNOB_fs_kwh, None)
        # set knobs
        if "wt_delta_perc" in self.knobs:
            # confirmed with Alicia and Jason on 8/11/2021 that light-weighting should occur before CdA adjustment
            self.weight_delta_percent_knob(wt_delta_perc_guess, optvehicle)
        if "CdA_perc_imp" in self.knobs:
            self.cda_percent_delta_knob(CdA_reduction_perc, optvehicle)
        if "fc_peak_eff" in self.knobs:
            self.fc_peak_eff_knob(fc_peak_eff_guess, optvehicle)
        if "fcMaxOutKw" in self.knobs:
            run_scenario.set_max_fuel_converter_kw(optvehicle, fc_max_out_kw_guess)
        if KNOB_fs_kwh in self.knobs:
            run_scenario.set_fuel_store_kwh(optvehicle, fs_kwh_guess)
        if "ess_max_kwh" in self.knobs:
            run_scenario.set_max_battery_kwh(optvehicle, max_ess_kwh_guess)
        if "mc_max_kw" in self.knobs:
            run_scenario.set_max_motor_kw(
                optvehicle, self.opt_scenario, max_motor_kw_guess
            )

        # enforce 0 <= cargo kg <= initial cargo kg for BEV and HEV optimizations
        if self.optimize_pt in [gl.BEV, gl.HEV]:
            run_scenario.limit_cargo_kg_for_moo_hev_bev(self.opt_scenario, optvehicle)

        # # TODO, this calculation and application of payload loss is still not finalized.
        # this also needs to be used for FCEV, would be nice if there were an FCEV veh_pt_type!!!
        # also it needs to have a Scenario File on/off activation

        assert len(x_dict) == 0, f"Unapplied knobs: {list(x_dict.keys())}"

        # calculate objectives
        get_accel_loaded = False
        get_grade = False
        obj_arr_F = []
        g6_acvhd, g125_acvhd, z60l_acvhd, z30l_acvhd, tco_acvhd, mpgge, range_achvd = [
            None
        ] * 7
        if GRADE6 in self.constr_list or GRADE125 in self.constr_list:
            get_grade = True
        if ACCEL30 in self.constr_list or ACCEL60 in self.constr_list:
            get_accel_loaded = True

        rs_sweep = run_scenario.vehicle_scenario_sweep(
            optvehicle,
            self.opt_scenario,
            designcycle,
            verbose=self.verbose,
            get_accel=False,  # don't want non-loaded accel values, for now
            get_accel_loaded=get_accel_loaded,
            get_gradeability=get_grade,
            write_tsv=write_tsv,
        )

        mpgge = rs_sweep["mpgge"]
        g6_acvhd, g125_acvhd = (
            rs_sweep["grade_6_mph_ach"],
            rs_sweep["grade_1_25_mph_ach"],
        )
        z60l_acvhd, z30l_acvhd = (
            rs_sweep["zero_to_60_loaded"],
            rs_sweep["zero_to_30_loaded"],
        )
        tco_acvhd = rs_sweep["tot_cost"]
        if optvehicle.veh_pt_type in [gl.BEV, gl.CONV, gl.HEV]:
            range_achvd = rs_sweep["primary_fuel_range_mi"]
        elif optvehicle.veh_pt_type == gl.PHEV:
            # need range from PHEV that is used to compare to target_range_mi
            range_achvd = rs_sweep["cd_aer_phev_range_mi"]

        if self.verbose:
            print(
                f"grade_6_mph_ach {g6_acvhd} grade_1_25_mph_ach {g125_acvhd} zero_to_60_loaded {z60l_acvhd} "
                f"zero_to_30_loaded {z30l_acvhd} tco {tco_acvhd} mpgge {mpgge} range {range_achvd}"
            )

        # PHEV fuel usage stats
        pct_fc_kwh = None
        phev_cd_fuel_used_kwh = None
        phev_cd_battery_used_kwh = None
        if (
            PHEV_MINIMIZE_FUEL_USE_OBJECTIVE in self.obj_list
            or PHEV_MINIMIZE_FUEL_USE_CONSTRAINT in self.constr_list
        ):
            phev_cd_fuel_used_kwh = rs_sweep["mpgge"]["cd_fuel_used_kwh_total"]
            phev_cd_battery_used_kwh = rs_sweep["mpgge"]["cd_battery_used_kwh"]
            pct_fc_kwh = round(
                phev_cd_fuel_used_kwh
                / (phev_cd_battery_used_kwh + phev_cd_fuel_used_kwh),
                2,
            )
        #                                                           #
        # ********************** objectives  ********************** #
        #                                                           #
        if TCO in self.obj_list:
            obj_arr_F.append(tco_acvhd)
        if PHEV_MINIMIZE_FUEL_USE_OBJECTIVE in self.obj_list:
            obj_arr_F.append(phev_cd_fuel_used_kwh)
        #                                                           #
        # ******************** end objectives  ******************** #
        #                                                           #

        #                                                           #
        # ********************** constraints ********************** #
        #
        #                                                           #

        self.accel_30_constraint.append(None)
        self.accel_60_constraint.append(None)
        self.grade_6_constraint.append(None)
        self.grade_1p25_constraint.append(None)
        self.range_constraint.append(None)
        self.grade_accel_overshoot_tol_constraint.append(None)
        self.c_rate_constraint.append(None)
        self.trace_miss_distance_percent_constraint_record.append(None)
        self.phev_min_fuel_use_prcnt_const_record.append(None)

        constraint_results_G = []
        # calculate constraint violations
        # speed at grade minus target should be negative when constraint is met
        if GRADE6 in self.constr_list:
            constraint_results_G.append(
                self.opt_scenario.min_speed_at_6pct_grade_in_5min_mph - g6_acvhd
            )
        if GRADE125 in self.constr_list:
            constraint_results_G.append(
                self.opt_scenario.min_speed_at_1p25pct_grade_in_5min_mph - g125_acvhd
            )
        if ACCEL60 in self.constr_list:
            # zero-to-speed time should minus max allowable (target value) should
            # be negative when constraint is met
            # 9 sec achvd - 10 sec target = -1
            constraint_results_G.append(
                z60l_acvhd - self.opt_scenario.max_time_0_to_60mph_at_gvwr_s
            )
        if ACCEL30 in self.constr_list:
            constraint_results_G.append(
                z30l_acvhd - self.opt_scenario.max_time_0_to_30mph_at_gvwr_s
            )

        # calculate limiting grade/accel requirement if all constraints met
        if ACCEL_GRADE_OVERSHOOT in self.constr_list:
            constr_perc = {}
            if GRADE125 in self.constr_list:
                # todo, -abs() for all of these?
                g125c = (
                    -(
                        g125_acvhd
                        - self.opt_scenario.min_speed_at_1p25pct_grade_in_5min_mph
                    )
                    / self.opt_scenario.min_speed_at_1p25pct_grade_in_5min_mph
                )
                constr_perc[GRADE125] = g125c
                self.grade_1p25_constraint[-1] = g125c
            if GRADE6 in self.constr_list:
                g6c = (
                    -(g6_acvhd - self.opt_scenario.min_speed_at_6pct_grade_in_5min_mph)
                    / self.opt_scenario.min_speed_at_6pct_grade_in_5min_mph
                )
                constr_perc[GRADE6] = g6c
                self.grade_6_constraint[-1] = g6c
            if ACCEL60 in self.constr_list:
                z60c = (
                    z60l_acvhd - self.opt_scenario.max_time_0_to_60mph_at_gvwr_s
                ) / self.opt_scenario.max_time_0_to_60mph_at_gvwr_s
                constr_perc[ACCEL60] = z60c
                self.accel_60_constraint[-1] = z60c
            if ACCEL30 in self.constr_list:
                z30c = (
                    z30l_acvhd - self.opt_scenario.max_time_0_to_30mph_at_gvwr_s
                ) / self.opt_scenario.max_time_0_to_30mph_at_gvwr_s
                constr_perc[ACCEL30] = z30c
                self.accel_30_constraint[-1] = z30c

            # if all constraints meet target, then find closest overshoot and ensure it's below tolerance
            if (len(constr_perc) > 0) & (
                np.array(list(constr_perc.values())) < 0
            ).all():
                # say largest negative percent is -.1 (closest to target, 10% exceeding)
                # tolerance is .05
                # -(-.1) - 0.05 = 0.05, >= 0, constraint is violated. Constraints are in bounds if they return < 0
                min_grade_accel_excess = (
                    -constr_perc[max(constr_perc, key=constr_perc.get)]
                    - self.grade_accel_overshoot_tol
                )
            else:
                min_grade_accel_excess = -1  # no constraint to satisfy
            constraint_results_G.append(min_grade_accel_excess)
            self.grade_accel_overshoot_tol_constraint[-1] = min_grade_accel_excess

        if RANGE in self.constr_list:
            # if you fall short of range target
            if self.range_overshoot_tol is None:
                range_mi_cv = (
                    self.opt_scenario.target_range_mi - range_achvd
                )  # pos return, failed
            else:
                if range_achvd <= self.opt_scenario.target_range_mi:
                    range_mi_cv = (
                        self.opt_scenario.target_range_mi - range_achvd
                    )  # pos return, failed
                else:
                    range_mi_cv = range_achvd - (
                        self.opt_scenario.target_range_mi
                        * (1 + self.range_overshoot_tol)
                    )

            constraint_results_G.append(range_mi_cv)
            self.range_constraint[-1] = range_mi_cv

        # c rate constraint
        if C_RATE in self.constr_list:
            self.c_rate_constraint[-1] = (
                self.mooadvancedvehicle.ess_max_kw / self.mooadvancedvehicle.ess_max_kwh
                - np.interp(
                    self.mooadvancedvehicle.ess_max_kwh,
                    # TODO, this 2D array needs to be an input
                    [1.0, 10.0, 188.0, 660.0],  # battery sizes kwh
                    [24.0, 12.0, 2.0, 0.7],  # c rates (kw/kwh)
                )
            )
            constraint_results_G.append(self.c_rate_constraint[-1])

        # # trace miss constraint
        if TRACE_MISS_DIST_PERCENT in self.constr_list:
            assert (
                self.opt_scenario.trace_miss_dist_percent > 0
                and self.opt_scenario.trace_miss_dist_percent < 1
            ), "scenario file input trace_miss_dist_percent must be decimal value greater than 0 and less than 1"
            cycle_records = rs_sweep["design_cycle_sim_drive_record"]
            max_dist_frac_result = max(
                sdr.trace_miss_dist_frac for sdr in cycle_records
            )
            # .1 -> 10%
            max_dist_frac_miss = self.opt_scenario.trace_miss_dist_percent
            constraint_results_G.append(max_dist_frac_result - max_dist_frac_miss)
            self.trace_miss_distance_percent_constraint_record[-1] = (
                max_dist_frac_result - max_dist_frac_miss
            )

        if PHEV_MINIMIZE_FUEL_USE_CONSTRAINT in self.constr_list:
            assert (
                self.opt_scenario.constraint_phev_minimize_fuel_use_percent > 0
            ), "scenario.constraint_phev_minimize_fuel_use_percent must be value > 0 and < 1"
            assert (
                self.opt_scenario.constraint_phev_minimize_fuel_use_percent < 1
            ), "scenario.constraint_phev_minimize_fuel_use_percent must be value > 0 and < 1"
            constraint_results_G.append(
                pct_fc_kwh - self.opt_scenario.constraint_phev_minimize_fuel_use_percent
            )
            self.phev_min_fuel_use_prcnt_const_record[-1] = (
                pct_fc_kwh - self.opt_scenario.constraint_phev_minimize_fuel_use_percent
            )

        #                                                           #
        # ******************** end constraints ******************** #
        #                                                           #

        # append reporting variables
        # obj vars
        self.r_tcos.append(tco_acvhd)
        self.r_cd_fc_kwh_percent.append(pct_fc_kwh)
        self.r_cd_fc_kwh_used.append(phev_cd_fuel_used_kwh)
        self.r_cd_elec_kwh_used.append(phev_cd_battery_used_kwh)

        self.r_grade_6s.append(g6_acvhd)
        self.r_grade_1p25s.append(g125_acvhd)
        self.r_accel_60l.append(z60l_acvhd)
        self.r_accel_30l.append(z30l_acvhd)
        self.r_fuel_efficiencies.append(mpgge)
        self.r_ranges.append(range_achvd)
        self.r_wt_delta_perc_guess.append(wt_delta_perc_guess)
        self.r_CdA_reduction_perc.append(CdA_reduction_perc)
        self.r_fc_peak_eff_guess.append(fc_peak_eff_guess)
        self.r_fc_max_out_kw_guess.append(fc_max_out_kw_guess)
        self.r_fs_kwh_guess.append(fs_kwh_guess)
        self.r_max_ess_kwh_guess.append(max_ess_kwh_guess)
        self.r_max_motor_kw_guess.append(max_motor_kw_guess)

        return np.array(obj_arr_F), np.array(constraint_results_G), rs_sweep

    def _evaluate(self, x: dict, out: dict, *args, **kwargs) -> None:
        """
        This method runs T3COProblem.get_objs() when running Pymoo optimization

        Args:
            x (dict): Dictionary containing optimization knobs
            out (dict): Dictionary containing TCO results for optimization runs
        """
        obj_arr_F, constr_arr, _ = self.get_objs(x)
        out["F"] = obj_arr_F

        if len(constr_arr) > 0:
            out["G"] = constr_arr

    def adjust_fc_peak_eff(
        self,
        fc_peak_eff: float,
        scenario: run_scenario.Scenario,
        optvehicle: fastsim.vehicle.Vehicle,
    ) -> None:
        """
        This method augments an advanced vehicle fc_eff_array based on new fc_peak_eff using baseline fc_eff_array


        Args:
            fc_peak_eff (float): Fuel converter peak efficiency override
            scenario (run_scenario.Scenario): Scenario object of current selection
            optvehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of optimization vehicle
        """
        old_peak_eff = scenario.origfc_eff_map.max()
        optvehicle.fc_eff_map = scenario.origfc_eff_map * (
            1 + (fc_peak_eff - old_peak_eff) / old_peak_eff
        )
        optvehicle.set_derived()

    # ------------------------------------ utility functions ------------------------------------
    def sweep_knob(
        self,
        knob: list,
        definition: int = 100,
        plot: bool = False,
        optres: float = None,
        **kwargs,
    ) -> list:
        """
        This method sweeps the optimization knob of vehicle from lbound to ubound, return TCO \
            plot optres to see if there's agreement from opt solution and your sweep

        Args:
            knob (list): list of knobs names for optimization
            definition (int, optional): Number of points. Defaults to 100.
            plot (bool, optional): if True, saves plot of bounds and TCOs. Defaults to False.
            optres (float, optional): Optimization resolution. Defaults to None.

        Returns:
            tcos (list): List of TCOs of length=definition
        """
        label = kwargs.get("label", "")
        sweep_knob_accel_test = kwargs.get("sweep_knob_accel_test", False)
        # save knobs and obj list
        temp_obj_list = self.obj_list
        self.knobs = [knob]
        self.obj_list = [TCO]
        lbound, ubound = self.knobs_bounds[knob]
        tcos = [self.get_objs([x])[0] for x in np.linspace(lbound, ubound, definition)]

        if sweep_knob_accel_test:
            opt_res_kw = np.inf
            if optres is not None:
                opt_res_kw = optres
            assert knob in [KNOB_mc_max_kw, KNOB_FCMAXKW]
            self.obj_list = [ACCEL30, ACCEL60]
            # objs.extend([z60l, z30l]) {(60, 30),...(60, 30)}
            thirty_mph_times = []
            sixty_mph_times = []
            largest_infeasible_30s_size = None
            failed_30_s_time = None
            largest_infeasible_60s_size = None
            failed_60_s_time = None
            for x in np.linspace(lbound, ubound, definition):
                z60l, z30l = self.get_objs([x])
                thirty_mph_times.append([z30l, x])
                sixty_mph_times.append([z60l, x])
            for res in sixty_mph_times:
                z60l, x = res
                if (
                    z60l > self.opt_scenario.max_time_0_to_60mph_at_gvwr_s
                    and x < opt_res_kw
                ):
                    largest_infeasible_60s_size = x
                    failed_60_s_time = z60l
            for res in sixty_mph_times:
                z30l, x = res
                if (
                    z30l > self.opt_scenario.max_time_0_to_30mph_at_gvwr_s
                    and x < opt_res_kw
                ):
                    largest_infeasible_30s_size = x
                    failed_30_s_time = z30l

        # todo, figure out how to get mpgge in the other y axis
        if plot:
            resdir = gl.MOO_KNOB_SWEEP_PLOTS_DIR
            if not resdir.exists():
                resdir.mkdir()
            ts = strftime("%Y-%m-%d_%H-%M-%S", gmtime())
            plt.figure(figsize=(7, 4))
            plt.ylabel("$TCO")
            plt.title(f"{label}_{ts}_{knob}")
            plt.xlabel(f"knob value for {definition} pts")
            plt.plot(
                np.linspace(lbound, ubound, definition),
                tcos,
                label=f"TCO from global optimum, sweeping {knob}",
            )
            if optres is not None:
                plt.vlines(
                    optres,
                    min(tcos),
                    max(tcos),
                    color="red",
                    label=f"MOO {knob} solution: {optres}",
                )
            if sweep_knob_accel_test:
                if largest_infeasible_30s_size is not None:
                    plt.vlines(
                        largest_infeasible_30s_size,
                        min(tcos),
                        max(tcos),
                        color="orange",
                        label=f"failed accel test 0 to 30 res: {failed_30_s_time} target: {self.opt_scenario.max_time_0_to_30mph_at_gvwr_s} {knob} size: {largest_infeasible_30s_size}",
                    )
                if largest_infeasible_60s_size is not None:
                    plt.vlines(
                        largest_infeasible_60s_size,
                        min(tcos),
                        max(tcos),
                        color="yellow",
                        label=f"failed accel test 0 to 60 res: {failed_60_s_time} target: {self.opt_scenario.max_time_0_to_60mph_at_gvwr_s} {knob} size: {largest_infeasible_30s_size}",
                    )
            plt.legend()
            plt.savefig(resdir / f"{knob}_{label}_{ts}.png")
            print(f"saving {knob} results to ", resdir)
            plt.show()
        self.knobs = self.knobs_bounds.keys()
        self.obj_list = temp_obj_list
        return tcos

    def get_tco_from_moo_advanced_result(self, x: dict) -> dict:
        """
        This method is a utility function to get detailed TCO information from optimized MOO result

        Args:
            x (dict): Dictionary containing optimization knobs - [max motor kw, battery kwh, drag coeff % improvement]

        Returns:
            out (dict): Dictionary containing TCO results for optimization runs
        """
        x_dict = {knob: x[self.knobs.index(knob)] for knob in self.knobs}
        wt_delta_perc_guess = x_dict.pop(KNOB_WTDELTAPERC, None)
        CdA_reduction_perc = x_dict.pop(KNOB_CDA, None)
        fc_peak_eff_guess = x_dict.pop(KNOB_FCPEAKEFF, None)
        fc_max_out_kw_guess = x_dict.pop(KNOB_FCMAXKW, None)
        max_ess_kwh_guess = x_dict.pop(KNOB_ess_max_kwh, None)
        max_motor_kw_guess = x_dict.pop(KNOB_mc_max_kw, None)
        fs_kwh_guess = x_dict.pop(KNOB_fs_kwh, None)

        print("MOO Final Solution:")
        if wt_delta_perc_guess is not None:
            print(KNOB_WTDELTAPERC.rjust(20, " "), f":{round(wt_delta_perc_guess, 4)}")
        if CdA_reduction_perc is not None:
            print(KNOB_CDA.rjust(20, " "), f":{round(CdA_reduction_perc, 4)}")
        if fc_peak_eff_guess is not None:
            print(KNOB_FCPEAKEFF.rjust(20, " "), f":{round(fc_peak_eff_guess, 4)}")
        if fc_max_out_kw_guess is not None:
            print(KNOB_FCMAXKW.rjust(20, " "), f":{round(fc_max_out_kw_guess, 4)}")
        if max_ess_kwh_guess is not None:
            print(KNOB_ess_max_kwh.rjust(20, " "), f":{round(max_ess_kwh_guess, 4)}")
        if max_motor_kw_guess is not None:
            print(KNOB_mc_max_kw.rjust(20, " "), f":{round(max_motor_kw_guess, 4)}")
        if fs_kwh_guess is not None:
            print(KNOB_fs_kwh.rjust(20, " "), f":{round(fs_kwh_guess, 4)}")

        _, _, out = self.get_objs(x, write_tsv=False)

        return out

    # ------------------------------------ end utility functions ------------------------------------


# TODO, needs refactor
class T3CODisplay():
    """
    This class contains the display object for Pymoo optimization printouts - pymoo.util.display.Display

    Args:
        Output (pymoo.util.display.output.Output): Pymoo minimize display object
    """

    def __init__(self, **kwargs) -> None:
        """
        This constructor initializes the pymoo.util.display.Display object
        """
        super().__init__(**kwargs)
        self.term = MultiObjectiveSpaceTermination()

    def _do(self, problem: T3COProblem, evaluator: float, algorithm: str) -> None:
        """
        This constructor creates the output printouts

        Args:
            problem (moo.T3COProblem): Pymoo optimization
            evaluator (float): evaluator
            algorithm (str): algorithm name
        """
        super()._do(problem, evaluator, algorithm)
        if problem.n_constr > 0:
            max_cv = np.max(algorithm.pop.get("G"))
            self.output.append("CV max", f"{max_cv:.3g}", width=10)
            # self.output.append("CV key",
            #     problem.constr_list[np.argmax(algorithm.pop.get('G')) % problem.n_constr] if max_cv > 0 else '-----',
            #     width=15
            # )
        if "tot_cost" in problem.obj_list:
            tco_start_idx = problem.obj_list.index("tot_cost")  # noqa: F841
            # tco_min_for_pop = min(algorithm.pop.get('F')[tco_start_idx::len(problem.obj_list)])[0]
            # self.output.append("min(TCO)", f"{tco_min_for_pop:.5g}")
        self.output.append("obj max", f"{np.max(algorithm.pop.get('F')):.5g}", width=10)
        self.output.append("n_nds", len(algorithm.opt), width=7)

        self.term.do_continue(algorithm)

        max_from, eps = "-", "-"
        if len(self.term.metrics) > 0:
            metric = self.term.metrics[-1]
            tol = self.term.tol
            delta_ideal, delta_nadir, delta_f = (
                metric["delta_ideal"],
                metric["delta_nadir"],
                metric["delta_f"],
            )
            if delta_ideal > tol:
                max_from = "ideal"
                eps = delta_ideal
            elif delta_nadir > tol:
                max_from = "nadir"
                eps = delta_nadir
            else:
                max_from = "f"
                eps = delta_f

        self.output.append("eps", eps)
        self.output.append("indicator", max_from)


def run_optimization(
    pop_size: int,
    n_max_gen: int,
    knobs_bounds: dict,
    vnum: int,
    x_tol: float,
    f_tol: float,
    nth_gen: int,
    n_last: int,
    algo: str,
    obj_list: list = None,
    config: run_scenario.Config = None,
    **kwargs,
) -> Tuple[pymoo.core.result.Result, T3COProblem, bool]:
    """
    This method creates and runs T3COProblem minimization

    Args:
        pop_size (int): Population size for optimization
        n_max_gen (int): maximum number of generations for optimization
        knobs_bounds (dict): Dictionary containing knobs and bounds
        vnum (int): vehicle selection number
        x_tol (float): tolerance in parameter space
        f_tol (float): tolerance in objective space
        nth_gen (int): number of generations to evaluate if convergence occurs
        n_last (int): number of generations to look back for termination
        algo (str): algorithm name
        obj_list (list, optional): list of objectives - TCO or PHEV_MINIMIZE_FUEL_USE_OBJECTIVE. Defaults to None.
        config (run_scenario.Config, optional): T3CO Config object containing analysis attributes and scenario attribute overrides. Defaults to None.

    Returns:
        res (pymoo.core.result.Result): Pymoo optimization result object
        problem (moo.T3COProblem): T3COProblem ElementwiseProblem object
        OPTIMIZATION_SUCCEEDED (bool): if True, pymoo.minimize succeeded
    """

    verbose = kwargs.pop("verbose", False)
    optimize_pt = kwargs.pop("optimize_pt", gl.BEV)
    return_least_infeasible = kwargs.pop("optimize_pt", False)
    skip_optimization = kwargs.pop("skip_optimization", False)

    if verbose:
        print("Running optimization.")

    if verbose:
        print(knobs_bounds)

    problem = T3COProblem(
        parallelization=("threads", 1),
        knobs_bounds=knobs_bounds,
        vnum=vnum,
        obj_list=obj_list,
        optimize_pt=optimize_pt,
        verbose=verbose,
        config=config,
        **kwargs,
    )

    if skip_optimization:
        return None, problem, None, None

    print(
        f"moo.run_optimization algo {algo}, x_tol {x_tol}, f_tol {f_tol}, nth_gen {nth_gen}, n_last {n_last}, n_max_gen {n_max_gen}, pop_size {pop_size}"
    )

    assert (
        algo in ALGORITHMS
    ), f"{algo} not in T3CO list of optimization algorithms {str(ALGORITHMS)}"

    if algo == ALGO_NSGA2:
        algorithm = NSGA2(
            pop_size=pop_size,
            eliminate_duplicates=True,
            # sampling=get_sampling(kwargs.pop('sampling', 'real_lhs'))
            sampling=kwargs.pop("sampling", LHS()),
        )
    elif algo == ALGO_NelderMead:
        print("moo.run_optimization Nelder Mead")
        algorithm = NelderMead()
    elif algo == ALGO_PatternSearch:
        print("moo.run_optimization PatternSearch")
        algorithm = PatternSearch()
    elif algo == ALGO_PSO:
        print("moo.run_optimization Particle Swarm")
        algorithm = PSO()
    # elif algo == 'LocalSearch':
    #     print('moo.run_optimization LocalSearch')
    #     algorithm = LocalSearch()
    t0 = time.time()

    termination = MODT(
        xtol=x_tol,
        ftol=f_tol,
        # n_last=n_last, these are expected in MODT... which is weird bc the docs seem to say it is
        # nth_gen=nth_gen,
        n_max_gen=n_max_gen,
        n_max_evals=None,
    )

    # this check no longer works now that kwargs are pass to T3COProblem and dict types are immutable
    # assert len(kwargs) == 0, f'Invalid kwargs: {list(kwargs.keys())}'
    try:
        res = minimize(
            problem,
            algorithm,
            termination=termination,
            seed=1,
            verbose=True,
            save_history=True,
            return_least_infeasible=return_least_infeasible,
            #    display=T3CODisplay()
        )
    except Exception:
        logging.exception(
            f"moo.run_optimization: Optimization errored out for algorithm {algo}"
        )
        res, problem = None, None
        return res, problem, EXCEPTION_THROWN

    t1 = time.time()
    print(f"\nElapsed time for optimization: {t1 - t0} s")
    if verbose:
        print("\nParameter pareto sets:")
    if res.X is None:
        print("moo.run_optimization: moo failed to converge")
        return res, problem, OPTIMIZATION_FAILED_TO_CONVERGE

    # res.X holds results of optimization
    # successful optimization could be 1D (one solution)
    # [1.0, 2.0, 3.0, 4.0, 5.0]
    # or nD, multiple solutions on a pareto front
    # [[1.0, 2.0, 3.0, 4.0, 5.0],
    # [ 6.0, 7.0, 8.0, 9.0, 10.0],
    # [ 11.0,12.0,13.0,14.0,15.0]]
    return res, problem, OPTIMIZATION_SUCCEEDED


# try:
#     ans = 1/0
# except BaseException as ex:
#     # Get current system exception
#     ex_type, ex_value, ex_traceback = sys.exc_info()

#     # Extract unformatter stack traces as tuples
#     trace_back = traceback.extract_tb(ex_traceback)

#     # Format stacktrace
#     stack_trace = list()

#     for trace in trace_back:
#         stack_trace.append("File : %s , Line : %d, Func.Name : %s, Message : %s" % (trace[0], trace[1], trace[2], trace[3]))

#     print("Exception type : %s " % ex_type.__name__)
#     print("Exception message : %s" %ex_value)
#     print("Stack trace : %s" %stack_trace)



"""
Testing module for TCO to ensure the TCO related results remain consistent over time with
previous model (Tech Targets c 2019)

Flow:
code reads TCO tests vehicles and scenarios one at a time. Generates TCO output dictionaries
and compares the results with the hard coded dictionaries below in comparison_dict.
"""
import numpy as np
from pathlib import Path
import ast

from t3co import run_scenario
from t3co.tco import tco_analysis
from t3co import Global as gl

# generated using Notebook from: https://github.nrel.gov/AVCI/FASTSim_TCO_Truck/blob/master/tco_baselining.ipynb
# and pasting attributes dict here
t2vehicle="""
{'selection': 1.0,
 'scenario_name': 'Conv 2020 tech,  750 mi range',
 'veh_pt_type': 1.0,
 'drag_coef': 0.546,
 'frontalAreaM2': 10.18,
 'glider_kg': 11776.0,
 'vehCgM': 0.46,
 'driveAxleWeightFrac': 0.8,
 'wheelBaseM': 2.28,
 'cargo_kg': 16329.34474,
 'vehOverrideKg': 0.0,
 'fs_max_kw': 5000.0,
 'fuelStorSecsToPeakPwr': 1.0,
 'fs_kwh': 3953.2751740000003,
 'fs_kwhPerKg': 9.875636446,
 'fc_max_kw': 340.0,
 'fc_eff_map': [0.05, 0.07, 0.12, 0.21, 0.26, 0.34, 0.38, 0.42, 0.45, 0.46, 0.46, 0.46],
 'fc_eff_type': 5.0,
 'fcAbsEffImpr': 0.0,
 'fuelConvSecsToPeakPwr': 6.0,
 'fuelConvBaseKg': 0.0,
 'fuelConvKwPerKg': 0.2745,
 'mcPwrOutPerc': [0.00, 0.02, 0.04, 0.06, 0.08, 0.10, 0.20, 0.40, 0.60, 0.80, 1.00],
 'largeBaselineEff': [0.83, 0.85, 0.87, 0.89, 0.90, 0.91, 0.93, 0.94, 0.94, 0.93, 0.92],
 'smallBaselineEff': [0.12, 0.16, 0.21, 0.29, 0.35, 0.42, 0.75, 0.92, 0.93, 0.93, 0.92],
 'modernMax': 0.95,
 'mc_max_kw': 0.0,
 'motorPeakEff': 0.93,
 'motorSecsToPeakPwr': 4.0,
 'stopStart': False,
 'mcPeKgPerKw': 0.333333333,
 'mcPeBaseKg': 0.0,
 'ess_max_kw': 0.0,
 'ess_max_kwh': 0.0,
 'essKgPerKwh': 4.7,
 'essBaseKg': 75.0,
 'essRoundTripEff': 0.97,
 'essLifeCoefA': 110.0,
 'essLifeCoefB': -0.6811,
 'wheelInertiaKgM2': 10.0,
 'numWheels': 18.0,
 'wheelRrCoef': 0.0061,
 'wheelRadiusM': 0.506,
 'wheelCoefOfFric': 0.6,
 'min_soc': 0.1,
 'max_soc': 0.95,
 'essDischgToFcMaxEffPerc': 0.0,
 'essChgToFcMaxEffPerc': 0.0,
 'maxAccelBufferMph': 60.0,
 'maxAccelBufferPercOfUseableSoc': 0.2,
 'percHighAccBuf': 0.0,
 'mphFcOn': 55.0,
 'kwDemandFcOn': 100.0,
 'altEff': 1.0,
 'chgEff': 0.86,
 'auxKw': 0.0,
 'forceAuxOnFC': 0.0,
 'transKg': 300.0,
 'transEff': 0.95,
 'comp_mass_multiplier': 1.2,
 'essToFuelOkError': 0.005,
 'maxRegen': 0.9,
 'valUddsMpgge': 0.0,
 'valHwyMpgge': 0.0,
 'valCombMpgge': 0.0,
 'valUddsKwhPerMile': 0.0,
 'valHwyKwhPerMile': 0.0,
 'valCombKwhPerMile': 0.0,
 'valCdRangeMi': 0.0,
 'valConst65MphKwhPerMile': 0.0,
 'valConst60MphKwhPerMile': 0.0,
 'valConst55MphKwhPerMile': 0.0,
 'valConst45MphKwhPerMile': 0.0,
 'valUnadjUddsKwhPerMile': 0.0,
 'valUnadjHwyKwhPerMile': 0.0,
 'val0To60Mph': 0.0,
 'valEssLifeMiles': 0.0,
 'valRangeMiles': 0.0,
 'valVehBaseCost': 0.0,
 'valMsrp': 0.0,
 'minFcTimeOn': 30,
 'idleFcKw': 0.0,
 'MaxRoadwayChgKw_Roadway': range(0, 6),
 'MaxRoadwayChgKw': [0, 0, 0, 0, 0, 0],
 'chargingOn': 0,
 'noElecSys': True,
 'noElecAux': True,
 'vehTypeSelection': 1,
 'fcEffArray': np.array([0.1       , 0.108     , 0.116     , 0.124     , 0.132     ,
        0.14      , 0.146     , 0.152     , 0.158     , 0.164     ,
        0.17      , 0.176     , 0.182     , 0.188     , 0.194     ,
        0.2       , 0.2024    , 0.2048    , 0.2072    , 0.2096    ,
        0.212     , 0.2144    , 0.2168    , 0.2192    , 0.2216    ,
        0.224     , 0.2264    , 0.2288    , 0.2312    , 0.2336    ,
        0.236     , 0.248     , 0.26      , 0.275     , 0.29      ,
        0.305     , 0.32      , 0.32875   , 0.3375    , 0.355     ,
        0.3725    , 0.39      , 0.395     , 0.4       , 0.405     ,
        0.41      , 0.41166667, 0.41333333, 0.415     , 0.41666667,
        0.41833333, 0.42      , 0.4195    , 0.419     , 0.4185    ,
        0.418     , 0.4175    , 0.417     , 0.4165    , 0.416     ,
        0.4155    , 0.415     , 0.4145    , 0.414     , 0.4135    ,
        0.413     , 0.4125    , 0.412     , 0.4115    , 0.411     ,
        0.4105    , 0.41      , 0.4085    , 0.407     , 0.4055    ,
        0.404     , 0.4025    , 0.401     , 0.3995    , 0.398     ,
        0.3965    , 0.395     , 0.3935    , 0.392     , 0.3905    ,
        0.389     , 0.3875    , 0.386     , 0.3845    , 0.383     ,
        0.3815    , 0.38      , 0.375     , 0.37      , 0.365     ,
        0.36      , 0.355     , 0.35      , 0.345     , 0.34      ]),
 'fcKwOutArray': np.array([  0.  ,   0.34,   0.68,   1.02,   1.36,   1.7 ,   2.04,   2.38,
          2.72,   3.06,   3.4 ,   3.74,   4.08,   4.42,   4.76,   5.1 ,
          5.44,   5.78,   6.12,   6.46,   6.8 ,   7.14,   7.48,   7.82,
          8.16,   8.5 ,   8.84,   9.18,   9.52,   9.86,  10.2 ,  11.9 ,
         13.6 ,  15.3 ,  17.  ,  18.7 ,  20.4 ,  22.1 ,  23.8 ,  27.2 ,
         30.6 ,  34.  ,  37.4 ,  40.8 ,  44.2 ,  47.6 ,  51.  ,  54.4 ,
         57.8 ,  61.2 ,  64.6 ,  68.  ,  71.4 ,  74.8 ,  78.2 ,  81.6 ,
         85.  ,  88.4 ,  91.8 ,  95.2 ,  98.6 , 102.  , 105.4 , 108.8 ,
        112.2 , 115.6 , 119.  , 122.4 , 125.8 , 129.2 , 132.6 , 136.  ,
        139.4 , 142.8 , 146.2 , 149.6 , 153.  , 156.4 , 159.8 , 163.2 ,
        166.6 , 170.  , 173.4 , 176.8 , 180.2 , 183.6 , 187.  , 190.4 ,
        193.8 , 197.2 , 200.6 , 204.  , 221.  , 238.  , 255.  , 272.  ,
        289.  , 306.  , 323.  , 340.  ]),
 'maxFcEffKw': np.array(68.),
 'fcMaxOutkW': np.array(340.),
 'mcKwInArray': np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]),
 'mcKwOutArray': np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]),
 'mcMaxElecInKw': 0.0,
 'regenA': 500.0,
 'regenB': 0.99,
 'veh_kg': 30432.05057387713,
 'cargoMassKg': 16329.34474,
 'curbMassKg': 14102.705833877131,
 'gvwrMassKg': 30432.05057387713,
 'ess_mass_kg': 0.0,
 'mc_mass_kg': 0.0,
 'fc_mass_kg': 1486.3387978142077,
 'fs_mass_kg': 480.36703606292315,
 'kwhPerMile': 5.864161359481092,
 'mpgge': 5.746772289189199}
"""

t2vehicle = ast.literal_eval(t2vehicle)

veh_no = 1  #
vocation = "Conv 2020 tech, 500 mi range"

gl.vocation_scenario = vocation
# gl.FASTSIM_INPUTS = gl.T2COBENCHMARKDATADIR / 't3cobenchmarkFASTSimInputs.csv'
# gl.OTHER_INPUTS = gl.T2COBENCHMARKDATADIR / 't3cobenchmarkOtherInputs.csv'

# load the generated file of vehicles, drive cycles, and tech targets
t3vehicle = run_scenario.get_vehicle(veh_no, gl.T2COBENCHMARKDATADIR / 't3cobenchmarkFASTSimInputs.csv', use_jit=False)
scenario, range_cyc = run_scenario.get_scenario_and_cycle(veh_no, gl.T2COBENCHMARKDATADIR / 't3cobenchmarkOtherInputs.csv', use_jit=False)
scenario.drive_cycle = r't3co_benchmarking_long_haul_cyc.csv'
cycpath = Path(r'C:\Users\gpayne\Documents\__old_projects\_Truck_FSIM_TCO\fastsim_src\cycles')
range_cyc = run_scenario.load_design_cycle_from_scenario(scenario, cyc_file_path=Path(cycpath))




def benchmark_vehicles():

    notmatch = {}
    for vspec, t3val in t3vehicle.__dict__.items():
        if vspec in t2vehicle:
            try:
                t2val = t2vehicle[vspec]

                if list in [type(t3val), type(t2val)]:
                    t3val = list(t3val)
                    t2val = list(t2val)

                if t2val != t3val:
                    notmatch[vspec] = {"t3val": t3val, "t2val": t2val}
                else:
                    print(f"{vspec} t3 {t3val} t2 {t2val}")

            except ValueError:
                pass


    assert notmatch == {}, notmatch

def benchmark_sim_outputs(t3out):

    # generated using Notebook from: https://github.nrel.gov/AVCI/FASTSim_TCO_Truck/blob/master/tco_baselining.ipynb
    # and pasting results and vars here
    t2output = {
        'Glider': 121918.9997,
        'Fuel converter': 31623.0,
        'Fuel Storage': 415.0938932700001,
        'Motor & power electronics': 0,
        'Plug': 0,
        'Battery': 0,
        'Battery replacement': 0,
        'Purchase tax': 0.0,
        'msrp': 153957.09359327,
        'tot_cost': 578491.5358170538,
        "disc_cost": 531496.4249639589,
        "mpgge": 6.737434503166163,
        "life_time_fuel_cost": 270534.442224,
        "discounted_life_time_fuel_cost": 240586.822343,
        }

    t3out["discounted_costs_df"].to_csv(gl.T2COBENCHMARKDATADIR / "t3co_discounted_costs_df_benchmark.csv")

    t3fuelagg = t3out["discounted_costs_df"].groupby(["Category"]).sum()
    t3_life_fuel = t3fuelagg.loc["Fuel", "Cost [$]"]
    t3_life_fuel_disc = t3fuelagg.loc["Fuel", "Discounted Cost [$]"]

    t3 = t3out['mpgge']
    t2 = t2output['mpgge']
    # mpgge t3co 6.737368952765226 != t2co 5.746772289189199 : %15.869680491069909
    print(f"mpgge t3co {t3} != t2co {t2} : %{abs(t3-t2)/((t3+t2)/2) * 100}")
    assert np.isclose(t3, t2)

    t2 = t2output["discounted_life_time_fuel_cost"]
    t3 = t3_life_fuel_disc
    print(f"discounted lifetime fuel cost t3co {t3} != t2co {t2} : %{abs(t3-t2)/((t3+t2)/2) * 100}")
    assert np.isclose(t3, t2)

    t2 = t2output["life_time_fuel_cost"]
    t3 = t3_life_fuel
    print(f"lifetime fuel cost t3co {t3} != t2co {t2} : %{abs(t3-t2)/((t3+t2)/2) * 100}")
    assert np.isclose(t3, t2)

    t3totcost = t3out['tot_cost']
    t2totcost = t2output['tot_cost']
    print(f"total TCO t3co {t3totcost} != t2co {t2totcost} : %{abs(t3totcost-t2totcost)/((t3totcost+t2totcost)/2) * 100}")
    assert np.isclose(t3totcost, t2totcost)


# hack, see if it makes mpgge align
t3vehicle.maxFcEffKw = np.array(68.)

out = run_scenario.rerun(t3vehicle, vocation, scenario, use_jit=False)
print("run scenario mpgge", out['mpgge'])

tot_cost, disc_cost, oppy_cost_set, discounted_costs_df, mpgge, veh_cost_set, sim_drive_record, veh_oper_cost_set,veh_opp_cost_set = tco_analysis.get_tco_of_vehicle(t3vehicle, range_cyc, scenario, out['mpgge'],use_jit=False)
print("manual mpgge", mpgge)
benchmark_vehicles()
benchmark_sim_outputs(out)

range_cyc = [(range_cyc, .3333)]*3
print("composite cycles test")
tot_cost, disc_cost, oppy_cost_set, discounted_costs_df, mpgge, veh_cost_set, sim_drive_record, veh_oper_cost_set, veh_opp_cost_set = tco_analysis.get_tco_of_vehicle(t3vehicle, range_cyc, scenario, out['mpgge'],use_jit=False)
print("manual mpgge", mpgge)
benchmark_vehicles()
benchmark_sim_outputs(out)
